@@ -3,9 +3,13 @@
 
 #undef DRIVER_OSS
 #undef DRIVER_ALSA
+#undef DRIVER_ALSA_050
 #undef DRIVER_ESD
+#undef DRIVER_SGI
 #undef USE_GNOME
 #undef NO_AUDIOFILE
+#undef NO_GASP
+#undef NO_ASM
 
 #undef ENABLE_NLS
 #undef HAVE_CATGETS
