@@ -2,7 +2,7 @@
 /*
  * The Real SoundTracker - main user interface handling
  *
- * Copyright (C) 1998-1999 Michael Krause
+ * Copyright (C) 1998-2001 Michael Krause
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -58,15 +58,15 @@
 #include "gui-settings.h"
 #include "file-operations.h"
 #include "playlist.h"
+#include "extspinbutton.h"
 
 int gui_playing_mode = 0;
+int notebook_current_page = NOTEBOOK_PAGE_FILE;
 GtkWidget *editing_toggle;
 GtkWidget *gui_curins_name, *gui_cursmpl_name;
 GtkWidget *mainwindow;
 ScopeGroup *scopegroup;
 
-static GSList *text_entries;
-
 static gint pipetag = -1;
 static GtkWidget *mainwindow_upper_hbox, *mainwindow_second_hbox;
 static GtkWidget *notebook;
@@ -76,27 +76,27 @@ static GtkAdjustment *adj_amplification, *adj_pitchbend;
 static GtkWidget *spin_jump, *curins_spin, *spin_octave;
 static Playlist *playlist;
 
-static void tempo_slider_changed (int value);
-static void bpm_slider_changed (int value);
+guint statusbar_context_id;
+GtkWidget *status_bar;
+GtkWidget *st_clock;
 
-static gui_subs_slider tempo_slider = {
-    N_("Tempo"), 1, 31, tempo_slider_changed, GUI_SUBS_SLIDER_SPIN_ONLY
+static void gui_tempo_changed (int value);
+static void gui_bpm_changed (int value);
+
+gui_subs_slider tempo_slider = {
+    N_("Tempo"), 1, 31, gui_tempo_changed, GUI_SUBS_SLIDER_SPIN_ONLY
 };
-static gui_subs_slider bpm_slider = {
-    "BPM", 32, 255, bpm_slider_changed, GUI_SUBS_SLIDER_SPIN_ONLY
+gui_subs_slider bpm_slider = {
+    "BPM", 32, 255, gui_bpm_changed, GUI_SUBS_SLIDER_SPIN_ONLY
 };
 
 static GdkColor gui_clipping_led_on, gui_clipping_led_off;
 static GtkWidget *gui_clipping_led;
 static gboolean gui_clipping_led_status;
 
-static int set_songpos_count = 0, startstop_count = 0;
-static double set_songpos_wait_time = -1.0;
-
 static int editing_pat = 0;
-static int songpos = 0;
-static int capture_keys = 1;
-static int notebook_current_page;
+
+static int gui_ewc_startstop = 0;
 
 /* gui event handlers */
 static void file_selected(GtkWidget *w, GtkFileSelection *fs);
@@ -105,9 +105,9 @@ static void current_instrument_name_changed(void);
 static void current_sample_changed(GtkSpinButton *spin);
 static void current_sample_name_changed(void);
 static int keyevent(GtkWidget *widget, GdkEventKey *event, gpointer data);
-static void spin_editpat_changed(GtkSpinButton *spin);
-static void spin_patlen_changed(GtkSpinButton *spin);
-static void spin_numchans_changed(GtkSpinButton *spin);
+static void gui_editpat_changed(GtkSpinButton *spin);
+static void gui_patlen_changed(GtkSpinButton *spin);
+static void gui_numchans_changed(GtkSpinButton *spin);
 static void notebook_page_switched(GtkNotebook *notebook, GtkNotebookPage *page, int page_num);
 static void gui_adj_amplification_changed(GtkAdjustment *adj);
 static void gui_adj_pitchbend_changed(GtkAdjustment *adj);
@@ -117,25 +117,26 @@ static void read_mixer_pipe(gpointer data, gint source, GdkInputCondition condit
 static void wait_for_player(void);
 static void play_song(void);
 static void play_pattern(void);
+static void play_current_pattern_row(void);
 
 /* gui initialization / helpers */
-static void text_entry_selected(void);
-static void text_entry_unselected(void);
-
 static void gui_enable(int enable);
 static void offset_current_pattern(int offset);
 static void offset_current_instrument(int offset);
 static void offset_current_sample(int offset);
-static void programlist_initialize(void);
+
+static void gui_auto_switch_page (void);
 
 static void
-gui_mixer_play_pattern (int p,
-			int pp)
+gui_mixer_play_pattern (int pattern,
+			int row,
+			int stop_after_row)
 {
     audio_ctlpipe_id i = AUDIO_CTLPIPE_PLAY_PATTERN;
     write(audio_ctlpipe, &i, sizeof(i));
-    write(audio_ctlpipe, &p, sizeof(p));
-    write(audio_ctlpipe, &pp, sizeof(pp));
+    write(audio_ctlpipe, &pattern, sizeof(pattern));
+    write(audio_ctlpipe, &row, sizeof(row));
+    write(audio_ctlpipe, &stop_after_row, sizeof(stop_after_row));
 }
 
 static void
@@ -167,6 +168,7 @@ gui_load_callback (gint reply,
 {
     if(reply == 0) {
 	gui_load_xm((gchar*)data);
+        gui_auto_switch_page();
     }
 }
 
@@ -175,8 +177,29 @@ gui_save_callback (gint reply,
 		   gpointer data)
 {
     if(reply == 0) {
-	XM_Save(xm, (gchar*)data);
-	xm->modified = 0;
+	statusbar_update(STATUS_SAVING_MODULE, TRUE);
+	if(XM_Save(xm, (gchar*)data, FALSE)) {
+	    xm->modified = 0;
+	    gui_auto_switch_page();
+	    statusbar_update(STATUS_MODULE_SAVED, FALSE);
+	} else {
+	    statusbar_update(STATUS_IDLE, FALSE);
+	}
+    }
+}
+
+static void
+gui_save_song_callback (gint reply,
+	                gpointer data)
+{
+    if(reply == 0) {
+	statusbar_update(STATUS_SAVING_SONG, TRUE);
+	if(XM_Save(xm, (gchar*)data, TRUE)) {
+	    gui_auto_switch_page();
+	    statusbar_update(STATUS_SONG_SAVED, FALSE);
+	} else {
+	    statusbar_update(STATUS_IDLE, FALSE);
+	}
     }
 }
 
@@ -231,6 +254,20 @@ file_selected (GtkWidget *w,
 	} else {
 	    gui_save_callback(0, fn);
 	}
+    } else if(fs == GTK_FILE_SELECTION(fileops_dialogs[DIALOG_SAVE_SONG_AS_XM])) {
+	FILE *f = fopen(fn, "r");
+
+	file_selection_save_path(fn, gui_settings.savesongasxm_path);
+
+	if(f != NULL) {
+	    fclose(f);
+	    gnome_app_ok_cancel_modal(GNOME_APP(mainwindow),
+				      _("Are you sure you want to overwrite the file?"),
+				      gui_save_song_callback,
+				      fn);
+	} else {
+	    gui_save_song_callback(0, fn);
+	}
     } else {
 	FILE *f = fopen(fn, "r");
 
@@ -303,7 +340,7 @@ gui_handle_standard_keys (int shift,
 			  int alt,
 			  guint32 keyval)
 {
-    gboolean handled = FALSE;
+    gboolean handled = FALSE, b;
 
     switch (keyval) {
     case GDK_F1 ... GDK_F7:
@@ -316,7 +353,29 @@ gui_handle_standard_keys (int shift,
 	if(ctrl) {
 	    gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin_jump), keyval - '0');
 	    handled = TRUE;
-	}
+        }
+        if(alt) {
+            switch(keyval)
+            {
+                case '1':
+                gtk_notebook_set_page(GTK_NOTEBOOK(notebook), 0);
+                break;
+                case '2':
+                gtk_notebook_set_page(GTK_NOTEBOOK(notebook), 1);
+                break;
+                case '3':
+                gtk_notebook_set_page(GTK_NOTEBOOK(notebook), 2);
+                break;
+                case '4':
+                gtk_notebook_set_page(GTK_NOTEBOOK(notebook), 3);
+                break;
+                case '5':
+                gtk_notebook_set_page(GTK_NOTEBOOK(notebook), 4);
+                break;
+                default:
+            }
+	    handled = TRUE;
+        }
 	break;
     case GDK_Left:
 	if(ctrl) {
@@ -367,19 +426,34 @@ gui_handle_standard_keys (int shift,
 	play_song();
 	handled = TRUE;
 	break;
+    case GDK_Menu:
+	play_current_pattern_row();
+	break;
     case ' ':
-	if(ctrl || alt)
-	    break;
-	if(!shift) {
-	    if(!GUI_ENABLED) {
-		gui_play_stop();
-	    } else {
-		gui_play_stop();
-		/* toggle editing mode */
-		gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(editing_toggle), !GUI_EDITING);
-	    }
-	    handled = TRUE;
-	}
+        if(ctrl || alt || shift)
+            break;
+	b = GUI_ENABLED;
+	gui_play_stop();
+	if(b) {
+            /* toggle editing mode (only if we haven't been in playing mode) */
+            gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(editing_toggle), !GUI_EDITING);
+            if (GTK_TOGGLE_BUTTON(editing_toggle)->active)
+		show_editmode_status();
+            else
+		statusbar_update(STATUS_IDLE, FALSE);
+        }
+	handled = TRUE;
+	break;
+    case GDK_Escape:
+        if(ctrl || alt || shift)
+            break;
+	/* toggle editing mode, even if we're in playing mode */
+	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(editing_toggle), !GUI_EDITING);
+        if (GTK_TOGGLE_BUTTON(editing_toggle)->active)
+	    show_editmode_status();
+        else
+	    statusbar_update(STATUS_IDLE, FALSE);
+	handled = TRUE;
 	break;
     }
 
@@ -400,8 +474,9 @@ keyevent (GtkWidget *widget,
     };
     gboolean pressed = (gboolean)data;
     gboolean handled = FALSE;
+    gboolean entry_focus = GTK_IS_ENTRY(GTK_WINDOW(mainwindow)->focus_widget);
 
-    if(capture_keys && GTK_WIDGET_VISIBLE(notebook)) {
+    if(!entry_focus && GTK_WIDGET_VISIBLE(notebook)) {
 	int shift = event->state & GDK_SHIFT_MASK;
 	int ctrl = event->state & GDK_CONTROL_MASK;
 	int alt = event->state & GDK_MOD1_MASK;
@@ -410,41 +485,53 @@ keyevent (GtkWidget *widget,
 	    handled = gui_handle_standard_keys(shift, ctrl, alt, event->keyval);
 	handled = handled || handle_page_keys[notebook_current_page](shift, ctrl, alt, event->keyval, pressed);
 
+	if(!handled) switch(event->keyval) {
+	    /* from gtk+-1.2.8's gtkwindow.c. These keypresses need to
+	       be stopped in any case. */
+	case GDK_Up:
+	case GDK_Down:
+	case GDK_Left:
+	case GDK_Right:
+	case GDK_KP_Up:
+	case GDK_KP_Down:
+	case GDK_KP_Left:
+	case GDK_KP_Right:
+	case GDK_Tab:
+	case GDK_ISO_Left_Tab:
+	    handled = TRUE;
+	}
+
 	if(handled) {
-	    if(pressed)
+	    if(pressed) {
 		gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
-	    else
+	    } else {
 		gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_release_event");
+	    }
 	}
     } else {
 	if(pressed) {
 	    switch(event->keyval) {
 	    case GDK_Tab:
 	    case GDK_Return:
+		gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "key_press_event");
 		gtk_window_set_focus(GTK_WINDOW(mainwindow), NULL);
-		capture_keys = 1;
 		break;
 	    }
 	}
     }
 
-    return 1;
+    return TRUE;
 }
 
 static void
 playlist_position_changed (Playlist *p,
 			   int newpos)
 {
-    if(newpos == songpos)
-	return;
-
-    songpos = newpos;
-
     if(gui_playing_mode) {
 	// This will only be executed when the user changes the song position manually
 	gui_mixer_set_songpos(newpos);
-	// See comment in audio.c::audio_mix() for why we do this
-	set_songpos_count++;
+	event_waiter_start(audio_songpos_ew);
+    } else {
 	gui_set_current_pattern(playlist_get_nth_pattern(p, newpos));
     }
 }
@@ -484,7 +571,7 @@ playlist_song_length_changed (Playlist *p,
 }
 
 static void
-spin_editpat_changed (GtkSpinButton *spin)
+gui_editpat_changed (GtkSpinButton *spin)
 {
     int n = gtk_spin_button_get_value_as_int(spin);
 
@@ -493,7 +580,7 @@ spin_editpat_changed (GtkSpinButton *spin)
 }
 
 static void
-spin_patlen_changed (GtkSpinButton *spin)
+gui_patlen_changed (GtkSpinButton *spin)
 {
     int n = gtk_spin_button_get_value_as_int(spin);
     XMPattern *pat = &xm->patterns[editing_pat];
@@ -507,7 +594,7 @@ spin_patlen_changed (GtkSpinButton *spin)
 }
 
 static void
-spin_numchans_changed (GtkSpinButton *spin)
+gui_numchans_changed (GtkSpinButton *spin)
 {
     int n = gtk_spin_button_get_value_as_int(spin);
 
@@ -523,23 +610,31 @@ spin_numchans_changed (GtkSpinButton *spin)
 }
 
 static void
-tempo_slider_changed (int value)
+gui_tempo_changed (int value)
 {
     audio_ctlpipe_id i;
     xm->tempo = value;
     xm_set_modified(1);
-    i = AUDIO_CTLPIPE_INIT_PLAYER;
+    if(gui_playing_mode) {
+	event_waiter_start(audio_tempo_ew);
+    }
+    i = AUDIO_CTLPIPE_SET_TEMPO;
     write(audio_ctlpipe, &i, sizeof(i));
+    write(audio_ctlpipe, &value, sizeof(value));
 }
 
 static void
-bpm_slider_changed (int value)
+gui_bpm_changed (int value)
 {
     audio_ctlpipe_id i;
     xm->bpm = value;
     xm_set_modified(1);
-    i = AUDIO_CTLPIPE_INIT_PLAYER;
+    if(gui_playing_mode) {
+	event_waiter_start(audio_bpm_ew);
+    }
+    i = AUDIO_CTLPIPE_SET_BPM;
     write(audio_ctlpipe, &i, sizeof(i));
+    write(audio_ctlpipe, &value, sizeof(value));
 }
 
 static void
@@ -576,36 +671,61 @@ notebook_page_switched (GtkNotebook *notebook,
     notebook_current_page = page_num;
 }
 
+/* gui_update_player_pos() is responsible for updating various GUI
+   features according to the current player position. ProTracker and
+   FastTracker scroll the patterns while the song is playing, and we
+   need the time-buffer and event-waiter interfaces here for correct
+   synchronization (we're called from
+   track-editor.c::tracker_timeout() which hands us time-correct data)
+
+   We have an ImpulseTracker-like editing mode as well ("asynchronous
+   editing"), which disables the scrolling, but still updates the
+   current song position spin buttons, for example. */
+
 void
-gui_update_player_pos (double songtime,
-		       int nsongpos,
-		       int npatpos)
+gui_update_player_pos (const audio_player_pos *p)
 {
     int m = xm_get_modified();
 
     if(gui_playing_mode == PLAYING_NOTE)
 	return;
 
-    /* This test prevents feedback which occurs when manually changing the song position;
-       if the audio thread doesn't follow immediately, the song position would be set back
-       to its old position some lines below... */
-    if(set_songpos_count != 0 || (set_songpos_wait_time != -1.0 && songtime < set_songpos_wait_time))
-	return;
-
-    set_songpos_wait_time = -1.0;
-
-    if(gui_playing_mode == PLAYING_SONG && nsongpos != songpos) {
-	// This will not be executed if the user has manually changed the song position
-	songpos = nsongpos;
-	playlist_set_position(playlist, songpos);
-	gui_set_current_pattern(xm->pattern_order_table[songpos]);
-    }
-    if(gui_playing_mode != PLAYING_NOTE) {
-	tracker_set_patpos(tracker, npatpos);
+    if(gui_playing_mode == PLAYING_SONG) {
+	if(event_waiter_ready(audio_songpos_ew, p->time)) {
+	    /* The following check prevents excessive calls of set_position() */
+	    if(p->songpos != playlist_get_position(playlist)) {
+		playlist_freeze_signals(playlist);
+		playlist_set_position(playlist, p->songpos);
+		playlist_thaw_signals(playlist);
+	    }
+	}
+	if(!ASYNCEDIT) {
+	    /* The following is a no-op if we're already in the right pattern */
+	    gui_set_current_pattern(xm->pattern_order_table[p->songpos]);
+	}
     }
 
-    if(notebook_current_page == 0) {
-	gdk_flush(); /* X drawing accumulates otherwise and makes pattern scrolling rather non-realtime :) */
+    if(!ASYNCEDIT) {
+	tracker_set_patpos(tracker, p->patpos);
+
+	if(notebook_current_page == 0) {
+	    /* Prevent accumulation of X drawing primitives */
+	    gdk_flush();
+	}
+    }
+
+    if(gui_settings.tempo_bpm_update) {
+	if(event_waiter_ready(audio_tempo_ew, p->time)) {
+	    tempo_slider.update_without_signal = TRUE;
+	    gui_subs_set_slider_value(&tempo_slider, p->tempo);
+	    tempo_slider.update_without_signal = FALSE;
+	}
+
+	if(event_waiter_ready(audio_bpm_ew, p->time)) {
+	    bpm_slider.update_without_signal = TRUE;
+	    gui_subs_set_slider_value(&bpm_slider, p->bpm);
+	    bpm_slider.update_without_signal = FALSE;
+	}
     }
 
     xm_set_modified(m);
@@ -616,12 +736,11 @@ gui_clipping_indicator_update (double songtime)
 {
     if(songtime < 0.0) {
 	gui_clipping_led_status = 0;
-	gtk_widget_draw(gui_clipping_led, NULL);
     } else {
 	audio_clipping_indicator *c = time_buffer_get(audio_clipping_indicator_tb, songtime);
 	gui_clipping_led_status = c && c->clipping;
-	gtk_widget_draw(gui_clipping_led, NULL);
     }
+    gtk_widget_draw(gui_clipping_led, NULL);
 }
 
 static void
@@ -645,48 +764,63 @@ read_mixer_pipe (gpointer data,
 
     switch(a) {
     case AUDIO_BACKPIPE_PLAYING_STOPPED:
-	if(startstop_count > 0) {
+        statusbar_update(STATUS_IDLE, FALSE);
+#ifdef USE_GNOME
+        gtk_clock_stop(GTK_CLOCK(st_clock));
+#endif
+
+        if(gui_ewc_startstop > 0) {
 	    /* can be equal to zero when the audio subsystem decides to stop playing on its own. */
-	    startstop_count--;
+	    gui_ewc_startstop--;
 	}
 	gui_playing_mode = 0;
 	scope_group_stop_updating(scopegroup);
 	tracker_stop_updating();
+	sample_editor_stop_updating();
 	gui_enable(1);
 	break;
 
     case AUDIO_BACKPIPE_PLAYING_STARTED:
+        statusbar_update(STATUS_PLAYING_SONG, FALSE);
+	/* fall through */
+
     case AUDIO_BACKPIPE_PLAYING_PATTERN_STARTED:
-	startstop_count--;
+        if(a == AUDIO_BACKPIPE_PLAYING_PATTERN_STARTED)
+	    statusbar_update(STATUS_PLAYING_PATTERN, FALSE);
+#ifdef USE_GNOME
+        gtk_clock_set_seconds(GTK_CLOCK(st_clock), 0);
+        gtk_clock_start(GTK_CLOCK(st_clock));
+#endif
+
+        gui_ewc_startstop--;
 	gui_playing_mode = (a == AUDIO_BACKPIPE_PLAYING_STARTED) ? PLAYING_SONG : PLAYING_PATTERN;
-	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(editing_toggle), 0);
+	if(!ASYNCEDIT) {
+	    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(editing_toggle), FALSE);
+	}
 	gui_enable(0);
 	scope_group_start_updating(scopegroup);
 	tracker_start_updating();
+	sample_editor_start_updating();
 	break;
 
     case AUDIO_BACKPIPE_PLAYING_NOTE_STARTED:
-	startstop_count--;
+	gui_ewc_startstop--;
 	if(!gui_playing_mode) {
 	    gui_playing_mode = PLAYING_NOTE;
 	    scope_group_start_updating(scopegroup);
 	    tracker_start_updating();
+	    sample_editor_start_updating();
 	}
 	break;
 
     case AUDIO_BACKPIPE_DRIVER_OPEN_FAILED:
-	startstop_count--;
-	break;
-
-    case AUDIO_BACKPIPE_SONGPOS_CONFIRM:
-	set_songpos_count--;
-	readpipe(source, &set_songpos_wait_time, sizeof(set_songpos_wait_time));
-	g_assert(set_songpos_count >= 0);
-	break;
+	gui_ewc_startstop--;
+        break;
 
     case AUDIO_BACKPIPE_ERROR_MESSAGE:
     case AUDIO_BACKPIPE_WARNING_MESSAGE:
-	readpipe(source, &x, sizeof(x));
+        statusbar_update(STATUS_IDLE, FALSE);
+        readpipe(source, &x, sizeof(x));
 	if(msgbuflen < x + 1) {
 	    g_free(msgbuf);
 	    msgbuf = g_new(char, x + 1);
@@ -694,13 +828,14 @@ read_mixer_pipe (gpointer data,
 	}
 	readpipe(source, msgbuf, x + 1);
 	if(a == AUDIO_BACKPIPE_ERROR_MESSAGE)
-	    gnome_error_dialog(msgbuf);
+            gnome_error_dialog(msgbuf);
 	else
-	    gnome_warning_dialog(msgbuf);
+            gnome_warning_dialog(msgbuf);
 	break;
 
     default:
 	fprintf(stderr, "\n\n*** read_mixer_pipe: unexpected backpipe id %d\n\n\n", a);
+	g_assert_not_reached();
 	break;
     }
 
@@ -712,8 +847,8 @@ wait_for_player (void)
 {
     struct pollfd pfd = { audio_backpipe, POLLIN, 0 };
 
-    startstop_count++;
-    while(startstop_count != 0) {
+    gui_ewc_startstop++;
+    while(gui_ewc_startstop != 0) {
 	g_return_if_fail(poll(&pfd, 1, -1) > 0);
 	read_mixer_pipe(NULL, audio_backpipe, 0);
     }
@@ -722,10 +857,12 @@ wait_for_player (void)
 static void
 play_song (void)
 {
-    int sp = songpos;
+    int sp = playlist_get_position(playlist);
     int pp = 0;
     audio_ctlpipe_id i = AUDIO_CTLPIPE_PLAY_SONG;
 
+    g_assert(xm != NULL);
+
     gui_play_stop();
 
     write(audio_ctlpipe, &i, sizeof(i));
@@ -740,10 +877,18 @@ static void
 play_pattern (void)
 {
     gui_play_stop();
-    gui_mixer_play_pattern(gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spin_editpat)), 0);
+    gui_mixer_play_pattern(gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spin_editpat)), 0, 0);
     wait_for_player();
 }
 
+static void
+play_current_pattern_row (void)
+{
+    gui_play_stop();
+    gui_mixer_play_pattern(gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spin_editpat)), tracker->patpos, 1);
+    gui_ewc_startstop++;
+}
+
 void
 gui_play_stop (void)
 {
@@ -761,7 +906,7 @@ gui_init_xm (int new_xm)
     write(audio_ctlpipe, &i, sizeof(i));
     tracker_reset(tracker);
     if(new_xm) {
-	programlist_initialize();
+	gui_playlist_initialize();
 	editing_pat = -1;
 	gui_set_current_pattern(xm->pattern_order_table[0]);
 	gtk_spin_button_set_value(GTK_SPIN_BUTTON(curins_spin), 1);
@@ -776,7 +921,7 @@ gui_init_xm (int new_xm)
     }
     gui_subs_set_slider_value(&tempo_slider, xm->tempo);
     gui_subs_set_slider_value(&bpm_slider, xm->bpm);
-    tracker_set_num_channels(tracker, xm->num_channels);
+    track_editor_set_num_channels(xm->num_channels);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin_numchans), xm->num_channels);
     scope_group_set_num_channels(scopegroup, xm->num_channels);
     xm_set_modified(m);
@@ -790,6 +935,7 @@ gui_free_xm (void)
     sample_editor_set_sample(NULL);
     tracker_set_pattern(tracker, NULL);
     XM_Free(xm);
+    xm = NULL;
 }
 
 void
@@ -807,13 +953,18 @@ gui_new_xm (void)
 void
 gui_load_xm (const char *filename)
 {
+    statusbar_update(STATUS_LOADING_MODULE, TRUE);
+
     gui_free_xm();
-    xm = XM_Load(filename);
+    xm = File_Load(filename);
 
     if(!xm) {
 	gui_new_xm();
-    } else
+	statusbar_update(STATUS_IDLE, FALSE);
+    } else {
 	gui_init_xm(1);
+	statusbar_update(STATUS_MODULE_LOADED, FALSE);
+    }
 }
 
 void
@@ -827,22 +978,28 @@ gui_play_note (int channel,
     write(audio_ctlpipe, &channel, sizeof(channel));
     write(audio_ctlpipe, &note, sizeof(note));
     write(audio_ctlpipe, &instrument, sizeof(instrument));
-    startstop_count++;
+    gui_ewc_startstop++;
 }
 
 void
-gui_play_note_full (int channel,
-		    int note,
+gui_play_note_full (unsigned channel,
+		    unsigned note,
 		    STSample *sample,
-		    int offset)
+		    guint32 offset,
+		    guint32 count)
 {
+    int x;
     audio_ctlpipe_id a = AUDIO_CTLPIPE_PLAY_NOTE_FULL;
+
+    g_assert(sizeof(int) == sizeof(unsigned));
+
     write(audio_ctlpipe, &a, sizeof(a));
     write(audio_ctlpipe, &channel, sizeof(channel));
     write(audio_ctlpipe, &note, sizeof(note));
     write(audio_ctlpipe, &sample, sizeof(sample));
-    write(audio_ctlpipe, &offset, sizeof(offset));
-    startstop_count++;
+    x = offset; write(audio_ctlpipe, &x, sizeof(x));
+    x = count; write(audio_ctlpipe, &x, sizeof(x));
+    gui_ewc_startstop++;
 }
 
 void
@@ -856,26 +1013,32 @@ gui_play_note_keyoff (int channel)
 static void
 gui_enable (int enable)
 {
-    gtk_widget_set_sensitive(vscrollbar, enable);
-    gtk_widget_set_sensitive(editing_toggle, enable);
-    gtk_widget_set_sensitive(spin_patlen, enable);
+    if(!ASYNCEDIT) {
+	gtk_widget_set_sensitive(vscrollbar, enable);
+	gtk_widget_set_sensitive(spin_patlen, enable);
+    }
     playlist_enable(playlist, enable);
 }
 
 void
 gui_set_current_pattern (int p)
 {
-    int m = xm_get_modified();
+    int m;
 
     if(editing_pat == p)
 	return;
 
+    m = xm_get_modified();
+
     editing_pat = p;
     tracker_set_pattern(tracker, &xm->patterns[p]);
     gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin_editpat), p);
     gui_update_pattern_data();
-    if(!GUI_ENABLED)
+
+    if(!GUI_ENABLED && !ASYNCEDIT) {
 	gui_mixer_set_pattern(p);
+    }
+
     xm_set_modified(m);
 }
 
@@ -945,7 +1108,10 @@ gui_get_current_octave_value (void)
 int
 gui_get_current_jump_value (void)
 {
-    return gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spin_jump));
+    if(!GUI_ENABLED && !ASYNCEDIT)
+	return 0;
+    else
+	return gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spin_jump));
 }
 
 int
@@ -986,41 +1152,6 @@ offset_current_sample (int offset)
     gui_set_current_sample(nv);
 }
 
-static void
-text_entry_selected (void)
-{
-    GSList *i;
-    GtkWidget *f;
-
-    if(capture_keys == 1) {
-	for(i = text_entries, f = GTK_WINDOW(mainwindow)->focus_widget; i; i = g_slist_next(i)) {
-	    if(f == i->data) {
-		capture_keys = 0;
-		break;
-	    }
-	}
-    }
-}
-
-static void
-text_entry_unselected (void)
-{
-    if(capture_keys == 0) {
-	capture_keys = 1;
-    }
-}
-
-void
-gui_add_text_entry_capture (GtkWidget *widget)
-{
-    gtk_signal_connect_after(GTK_OBJECT(widget), "focus_in_event",
-			     GTK_SIGNAL_FUNC(text_entry_selected), NULL);
-    gtk_signal_connect_after(GTK_OBJECT(widget), "focus_out_event",
-			     GTK_SIGNAL_FUNC(text_entry_unselected), NULL);
-
-    text_entries = g_slist_prepend(text_entries, widget);
-}
-
 void
 gui_get_text_entry (int length,
 		    void(*changedfunc)(),
@@ -1033,19 +1164,16 @@ gui_get_text_entry (int length,
     gtk_signal_connect_after(GTK_OBJECT(thing), "changed",
 			     GTK_SIGNAL_FUNC(changedfunc), NULL);
 
-    gui_add_text_entry_capture(thing);
-
     *widget = thing;
 }
 
-static void
-programlist_initialize (void)
+void
+gui_playlist_initialize (void)
 {
     int i;
 
     playlist_freeze_signals(playlist);
     playlist_freeze(playlist);
-    songpos = 0;
     playlist_set_position(playlist, 0);
     playlist_set_length(playlist, xm->song_length);
     for(i = 0; i < xm->song_length; i++) {
@@ -1084,12 +1212,23 @@ gui_go_to_fileops_page (void)
 			  0);
 }
 
+void
+gui_auto_switch_page (void)
+{
+    if(gui_settings.auto_switch)
+        gtk_notebook_set_page(GTK_NOTEBOOK(notebook),
+                              1);
+}
+
 int
 gui_init (int argc,
 	  char *argv[])
 {
     GtkWidget *thing, *mainvbox, *table, *hbox, *frame, *mainvbox0;
     GdkColormap *colormap;
+#ifdef USE_GNOME
+    GtkWidget *dockitem;
+#endif
 
 #ifdef USE_GNOME
     gnome_init("SoundTracker", VERSION, argc, argv);
@@ -1099,8 +1238,6 @@ gui_init (int argc,
 
     pipetag = gdk_input_add(audio_backpipe, GDK_INPUT_READ, read_mixer_pipe, NULL);
 
-    text_entries = g_slist_alloc();
-
 #ifdef USE_GNOME
     mainwindow = gnome_app_new("SoundTracker", "SoundTracker " VERSION);
 #else
@@ -1118,6 +1255,8 @@ gui_init (int argc,
     fileops_dialogs[DIALOG_SAVE_MOD_AS_WAV] = file_selection_create(_("Render module as WAV..."), file_selected);
     gtk_file_selection_set_filename(GTK_FILE_SELECTION(fileops_dialogs[DIALOG_SAVE_MOD_AS_WAV]), gui_settings.savemodaswav_path);
 #endif
+    fileops_dialogs[DIALOG_SAVE_SONG_AS_XM] = file_selection_create(_("Save song as XM..."), file_selected);
+    gtk_file_selection_set_filename(GTK_FILE_SELECTION(fileops_dialogs[DIALOG_SAVE_SONG_AS_XM]), gui_settings.savesongasxm_path);
 
     mainvbox0 = gtk_vbox_new(FALSE, 0);
     gtk_container_border_width(GTK_CONTAINER(mainvbox0), 0);
@@ -1129,7 +1268,7 @@ gui_init (int argc,
     gtk_widget_show(mainvbox0);
 
     menubar_create(mainwindow, mainvbox0);
-
+	
     mainvbox = gtk_vbox_new(FALSE, 4);
     gtk_container_border_width(GTK_CONTAINER(mainvbox), 4);
     gtk_box_pack_start(GTK_BOX(mainvbox0), mainvbox, TRUE, TRUE, 0);
@@ -1204,10 +1343,11 @@ gui_init (int argc,
 
     add_empty_hbox(hbox);
 
-    spin_numchans = gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(8, 2, 32, 2.0, 8.0, 0.0)), 0, 0);
+    spin_numchans = extspinbutton_new(GTK_ADJUSTMENT(gtk_adjustment_new(8, 2, 32, 2.0, 8.0, 0.0)), 0, 0);
+    extspinbutton_disable_size_hack(EXTSPINBUTTON(spin_numchans));
     gtk_box_pack_start(GTK_BOX(hbox), spin_numchans, FALSE, TRUE, 0);
     gtk_signal_connect(GTK_OBJECT(spin_numchans), "changed",
-		       GTK_SIGNAL_FUNC(spin_numchans_changed), NULL);
+		       GTK_SIGNAL_FUNC(gui_numchans_changed), NULL);
     gtk_widget_show(spin_numchans);
 
     hbox = gtk_hbox_new(FALSE, 4);
@@ -1220,11 +1360,12 @@ gui_init (int argc,
 
     add_empty_hbox(hbox);
 
-    spin_editpat = gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 255, 1.0, 10.0, 0.0)), 0, 0);
+    spin_editpat = extspinbutton_new(GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 255, 1.0, 10.0, 0.0)), 0, 0);
+    extspinbutton_disable_size_hack(EXTSPINBUTTON(spin_editpat));
     gtk_box_pack_start(GTK_BOX(hbox), spin_editpat, FALSE, TRUE, 0);
     gtk_widget_show(spin_editpat);
     gtk_signal_connect(GTK_OBJECT(spin_editpat), "changed",
-		       GTK_SIGNAL_FUNC(spin_editpat_changed), NULL);
+		       GTK_SIGNAL_FUNC(gui_editpat_changed), NULL);
     
     hbox = gtk_hbox_new(FALSE, 4);
     gtk_table_attach_defaults(GTK_TABLE(table), hbox, 1, 2, 4, 5);
@@ -1236,10 +1377,11 @@ gui_init (int argc,
 
     add_empty_hbox(hbox);
 
-    spin_patlen = gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(64, 1, 256, 1.0, 16.0, 0.0)), 0, 0);
+    spin_patlen = extspinbutton_new(GTK_ADJUSTMENT(gtk_adjustment_new(64, 1, 256, 1.0, 16.0, 0.0)), 0, 0);
+    extspinbutton_disable_size_hack(EXTSPINBUTTON(spin_patlen));
     gtk_box_pack_start(GTK_BOX(hbox), spin_patlen, FALSE, TRUE, 0);
     gtk_signal_connect(GTK_OBJECT(spin_patlen), "changed",
-		       GTK_SIGNAL_FUNC(spin_patlen_changed), NULL);
+		       GTK_SIGNAL_FUNC(gui_patlen_changed), NULL);
     gtk_widget_show(spin_patlen);
 
     /* Scopes Group or Instrument / Sample Listing */
@@ -1326,7 +1468,8 @@ gui_init (int argc,
     gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, TRUE, 0);
     gtk_widget_show(thing);
 
-    spin_octave = gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(5.0, 0.0, 6.0, 1.0, 1.0, 0.0)), 0, 0);
+    spin_octave = extspinbutton_new(GTK_ADJUSTMENT(gtk_adjustment_new(5.0, 0.0, 6.0, 1.0, 1.0, 0.0)), 0, 0);
+    extspinbutton_disable_size_hack(EXTSPINBUTTON(spin_octave));
     gtk_box_pack_start(GTK_BOX(hbox), spin_octave, FALSE, TRUE, 0);
     gtk_widget_show(spin_octave);
 
@@ -1334,7 +1477,8 @@ gui_init (int argc,
     gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, TRUE, 0);
     gtk_widget_show(thing);
 
-    spin_jump = gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(1.0, 0.0, 8.0, 1.0, 1.0, 0.0)), 0, 0);
+    spin_jump = extspinbutton_new(GTK_ADJUSTMENT(gtk_adjustment_new(1.0, 0.0, 16.0, 1.0, 1.0, 0.0)), 0, 0);
+    extspinbutton_disable_size_hack(EXTSPINBUTTON(spin_jump));
     gtk_box_pack_start(GTK_BOX(hbox), spin_jump, FALSE, TRUE, 0);
     gtk_widget_show(spin_jump);
 
@@ -1342,7 +1486,8 @@ gui_init (int argc,
     gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, TRUE, 0);
     gtk_widget_show(thing);
 
-    curins_spin = gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(1.0, 1.0, 128.0, 1.0, 16.0, 0.0)), 0, 0);
+    curins_spin = extspinbutton_new(GTK_ADJUSTMENT(gtk_adjustment_new(1.0, 1.0, 128.0, 1.0, 16.0, 0.0)), 0, 0);
+    extspinbutton_disable_size_hack(EXTSPINBUTTON(curins_spin));
     gtk_box_pack_start(GTK_BOX(hbox), curins_spin, FALSE, TRUE, 0);
     gtk_widget_show(curins_spin);
     gtk_signal_connect (GTK_OBJECT(curins_spin), "changed",
@@ -1357,7 +1502,8 @@ gui_init (int argc,
     gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, TRUE, 0);
     gtk_widget_show(thing);
 
-    cursmpl_spin = gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(0.0, 0.0, 15.0, 1.0, 4.0, 0.0)), 0, 0);
+    cursmpl_spin = extspinbutton_new(GTK_ADJUSTMENT(gtk_adjustment_new(0.0, 0.0, 15.0, 1.0, 4.0, 0.0)), 0, 0);
+    extspinbutton_disable_size_hack(EXTSPINBUTTON(cursmpl_spin));
     gtk_box_pack_start(GTK_BOX(hbox), cursmpl_spin, FALSE, TRUE, 0);
     gtk_widget_show(cursmpl_spin);
     gtk_signal_connect (GTK_OBJECT(cursmpl_spin), "changed",
@@ -1389,23 +1535,51 @@ gui_init (int argc,
 			  1);
     notebook_current_page = 1;
 
-    /* status bar */
-/*
-    hbox = gtk_hbox_new(FALSE, 4);
-    gtk_box_pack_start(GTK_BOX(mainvbox), hbox, FALSE, TRUE, 0);
+    /* Status Bar */
+
+#ifdef USE_GNOME
+    dockitem = gnome_dock_item_new("Status Bar", (GNOME_DOCK_ITEM_BEH_EXCLUSIVE | GNOME_DOCK_ITEM_BEH_NEVER_VERTICAL));
+    gnome_app_add_dock_item(GNOME_APP(mainwindow), GNOME_DOCK_ITEM(dockitem), GNOME_DOCK_BOTTOM, 0, 0, 0);
+    gtk_widget_show(dockitem);
+
+    hbox = gtk_hbox_new(FALSE, 2);
+    gtk_container_border_width(GTK_CONTAINER(hbox), 2);
+    gtk_container_add(GTK_CONTAINER(dockitem), hbox);
     gtk_widget_show(hbox);
-    
-    add_empty_hbox(hbox);
-    thing = gtk_vseparator_new();
-    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, TRUE, 0);
+
+    status_bar = gnome_appbar_new (FALSE, TRUE, GNOME_PREFERENCES_NEVER);
+    gtk_widget_show (status_bar);
+    gtk_box_pack_start (GTK_BOX (hbox), status_bar, TRUE, TRUE, 0);
+    gtk_widget_set_usize (status_bar, 300 , 20); /* so that it doesn't vanish when undocked */
+
+    thing = gtk_frame_new (NULL);
+    gtk_widget_show (thing);
+    gtk_box_pack_start (GTK_BOX (hbox), thing, FALSE, FALSE, 0);
+    gtk_widget_set_usize (thing, 48, 20);
+    gtk_frame_set_shadow_type (GTK_FRAME (thing), GTK_SHADOW_IN);
+
+    st_clock = gtk_clock_new (GTK_CLOCK_INCREASING);
+    gtk_widget_show (st_clock);
+    gtk_container_add (GTK_CONTAINER (thing), st_clock);
+    gtk_widget_set_usize (st_clock, 48, 20);
+    gtk_clock_set_format (GTK_CLOCK (st_clock), _("%M:%S"));
+    gtk_clock_set_seconds(GTK_CLOCK (st_clock), 0);
+
+    gnome_appbar_set_status(GNOME_APPBAR(status_bar), WELCOME_MESSAGE);
+#else
+    thing = gtk_hbox_new(FALSE, 1);
+    gtk_box_pack_start(GTK_BOX(mainvbox), thing, FALSE, TRUE, 0);
     gtk_widget_show(thing);
 
-    thing = gui_button (mainwindow, GNOME_STOCK_BUTTON_UP,
-			0, NULL, NULL, 0);
-    gtk_box_pack_start (GTK_BOX (hbox), thing, 0, 0, 0);
-    gtk_widget_show (thing);
-//    add_tip (wid, "Link/DeLink this tab");
-*/
+    status_bar = gtk_statusbar_new();
+    gtk_box_pack_start(GTK_BOX (thing), status_bar, TRUE, TRUE, 0);
+    gtk_widget_show(status_bar);
+    gtk_widget_set_usize(status_bar, -2, 20);
+
+    statusbar_context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(status_bar), "ST Statusbar");
+    gtk_statusbar_push(GTK_STATUSBAR(status_bar), statusbar_context_id, WELCOME_MESSAGE);
+#endif
+    
     /* capture all key presses */
     gtk_widget_add_events(GTK_WIDGET(mainwindow), GDK_KEY_RELEASE_MASK);
     gtk_signal_connect(GTK_OBJECT(mainwindow), "key_press_event", GTK_SIGNAL_FUNC(keyevent), (gpointer)1);
