@@ -1,8 +1,7 @@
-
 /*
  * The Real SoundTracker - GUI (menu bar)
  *
- * Copyright (C) 1999 Michael Krause
+ * Copyright (C) 1999-2001 Michael Krause
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,6 +20,8 @@
 
 #include <config.h>
 
+#include <string.h>
+
 #include "i18n.h"
 #include "menubar.h"
 #include "gui.h"
@@ -33,21 +34,26 @@
 #include "scope-group.h"
 #include "track-editor.h"
 #include "audioconfig.h"
-#include "effectsconfig.h"
 #include "gui-settings.h"
 #include "tips-dialog.h"
 #include "transposition.h"
 #include "cheat-sheet.h"
 #include "file-operations.h"
 #include "instrument-editor.h"
+#include "tracker-settings.h"
+#include "midi-settings.h"
 
 #ifdef USE_GNOME
 #include <gnome.h>
 #else
 static GtkItemFactory *item_factory;
 static GtkAccelGroup *accel_group;
+#ifdef ENABLE_NLS
 static int all_translated;
 #endif
+#endif
+
+static gboolean mark_mode_toggle_ignore = FALSE;
 
 extern ScopeGroup *scopegroup;
 
@@ -69,7 +75,7 @@ about_dialog (void)
     const gchar *authors[] = {"Michael Krause <m.krause@tu-harburg.de>", NULL};
     GtkWidget *about = gnome_about_new("SoundTracker",
 				       VERSION,
-				       "Copyright (C) 1998-1999 Michael Krause",
+				       "Copyright (C) 1998-2001 Michael Krause",
 				       authors,
 				       "Includes OpenCP player from Niklas Beisert and Tammo Hinrichs.",
 				       NULL);
@@ -84,7 +90,7 @@ about_dialog (void)
     gtk_window_position (GTK_WINDOW(about), GTK_WIN_POS_CENTER);
     gtk_window_set_title(GTK_WINDOW(about), "About SoundTracker");
 
-    label = gtk_label_new("SoundTracker " VERSION "\n\nCopyright (C) 1998-1999 Michael Krause\n<m.krause@tu-harburg.de>"
+    label = gtk_label_new("SoundTracker " VERSION "\n\nCopyright (C) 1998-2001 Michael Krause\n<m.krause@tu-harburg.de>"
 			  "\n\nIncludes OpenCP player from Niklas Beisert and Tammo Hinrichs.");
     gtk_container_border_width(GTK_CONTAINER(GTK_DIALOG(about)->vbox), 10);
     gtk_box_pack_start(GTK_BOX(GTK_DIALOG(about)->vbox), label, TRUE, TRUE, 10);
@@ -178,6 +184,15 @@ _FUNCMACRO_(menubar_scopes_toggled)
     scope_group_enable_scopes(scopegroup, gui_settings.gui_display_scopes);
 }
 
+static void
+_FUNCMACRO_(menubar_mark_mode_toggled)
+{
+    if(!mark_mode_toggle_ignore) {
+	gboolean state = GTK_CHECK_MENU_ITEM(widget)->active;
+	tracker_mark_selection(tracker, state);
+    }
+}
+
 static void
 _FUNCMACRO_(menubar_save_settings_on_exit_toggled)
 {
@@ -190,6 +205,10 @@ menubar_save_settings_now (void)
     gui_settings_save_config();
     keys_save_config();
     audioconfig_save_config();
+    trackersettings_write_settings();
+#if defined(DRIVER_ALSA_050) && defined(USE_GNOME)
+    midi_save_config();
+#endif
 }
 
 static void
@@ -199,21 +218,60 @@ menubar_handle_edit_menu (void *p,
     Tracker *t = tracker;
 
     static void (* const functions[]) (Tracker *) = {
-	track_editor_cut_pattern,
-	track_editor_copy_pattern,
-	track_editor_paste_pattern,
-	track_editor_cut_track,
-	track_editor_copy_track,
-	track_editor_paste_track,
-	track_editor_mark_selection,
-	track_editor_cut_selection,
-	track_editor_copy_selection,
-	track_editor_paste_selection,
-	track_editor_insert_track,
-	track_editor_delete_track
+	track_editor_cut_pattern,                       // 0
+	track_editor_copy_pattern,                      // 1
+	track_editor_paste_pattern,                     // 2
+	track_editor_cut_track,                         // 3
+        track_editor_copy_track,                        // 4
+        track_editor_paste_track,                       // 5
+	track_editor_insert_track,                      // 6
+        track_editor_delete_track,                      // 7
+	track_editor_interpolate_fx,                    // 8
+        track_editor_clear_mark_selection,              // 9
+        track_editor_cut_selection,                     // 10
+        track_editor_copy_selection,                    // 11
+        track_editor_paste_selection,                   // 12
+	track_editor_kill_notes_track,                   // 13
+        NULL,                                           // 14
+        NULL                                            // 15
     };
 
-    functions[a](t);
+#ifdef USE_GNOME
+    GtkWidget *active;
+    GtkMenu *menu = (GtkMenu *)a;
+    /* Check if it's call from popup */
+    if(a > sizeof(functions)/sizeof(functions[0]) && GTK_IS_WIDGET(menu)) {
+	/* Get active menu */
+	active = gtk_object_get_data (GTK_OBJECT (menu), "gnome_popup_menu_active_item");
+	/* Get user_data for this menu */
+	a = (gint)gtk_object_get_data (GTK_OBJECT (active), GNOMEUIINFO_KEY_UIDATA);
+    }
+#endif
+
+    switch(a) {
+    case 14:
+        track_editor_cmd_mvalue(t, TRUE);   /* increment CMD value */
+        break;
+    case 15:
+        track_editor_cmd_mvalue(t, FALSE);  /* decrement CMD value */
+        break;
+
+    default:
+        functions[a](t);
+        break;
+    }
+}
+
+static void
+menubar_settings_tracker_next_font (void)
+{
+    trackersettings_cycle_font_forward(TRACKERSETTINGS(trackersettings));
+}
+
+static void
+menubar_settings_tracker_prev_font (void)
+{
+    trackersettings_cycle_font_backward(TRACKERSETTINGS(trackersettings));
 }
 
 #ifdef USE_GNOME
@@ -229,6 +287,8 @@ static GnomeUIInfo file_menu[] = {
 #define SAVE_MOD_AS_WAV_POSITION 3
     { GNOME_APP_UI_ITEM, N_("Save Module as _WAV..."), NULL, fileops_open_dialog, (gpointer)DIALOG_SAVE_MOD_AS_WAV, NULL,
       GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_SAVE_AS, 0, 0, NULL },
+    { GNOME_APP_UI_ITEM, N_("Save XM without samples..."), NULL, fileops_open_dialog, (gpointer)DIALOG_SAVE_SONG_AS_XM, NULL,
+      GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_SAVE_AS, 0, 0, NULL },
 
     GNOMEUIINFO_SEPARATOR,
 
@@ -267,26 +327,36 @@ static GnomeUIInfo edit_track_menu[] = {
       GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_COPY, GDK_F4, GDK_SHIFT_MASK, NULL },
     { GNOME_APP_UI_ITEM, N_("_Paste"), NULL, menubar_handle_edit_menu, (gpointer)5, NULL,
       GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PASTE, GDK_F5, GDK_SHIFT_MASK, NULL },
-    { GNOME_APP_UI_ITEM, N_("_Insert"), NULL, menubar_handle_edit_menu, (gpointer)10, NULL,
+    { GNOME_APP_UI_ITEM, N_("_Kill notes"), NULL, menubar_handle_edit_menu, (gpointer)13, NULL,
+      GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_NEW, GDK_K, GDK_CONTROL_MASK, NULL },
+    { GNOME_APP_UI_ITEM, N_("_Insert"), NULL, menubar_handle_edit_menu, (gpointer)6, NULL,
       GNOME_APP_PIXMAP_NONE, 0, 0, 0, NULL },
-    { GNOME_APP_UI_ITEM, N_("_Delete"), NULL, menubar_handle_edit_menu, (gpointer)11, NULL,
+    { GNOME_APP_UI_ITEM, N_("_Delete"), NULL, menubar_handle_edit_menu, (gpointer)7, NULL,
       GNOME_APP_PIXMAP_NONE, 0, 0, 0, NULL },
-
+    { GNOME_APP_UI_ITEM, N_("Increment cmd value"), NULL, menubar_handle_edit_menu, (gpointer)14, NULL,
+      GNOME_APP_PIXMAP_NONE, GNOME_STOCK_MENU_NEW, '=', GDK_CONTROL_MASK, NULL },
+    { GNOME_APP_UI_ITEM, N_("Decrement cmd value"), NULL, menubar_handle_edit_menu, (gpointer)15, NULL,
+      GNOME_APP_PIXMAP_NONE, GNOME_STOCK_MENU_NEW, '-', GDK_CONTROL_MASK, NULL },
 
     GNOMEUIINFO_END
 };
 
 static GnomeUIInfo edit_selection_menu[] = {
-    { GNOME_APP_UI_ITEM, N_("_Mark"), NULL, menubar_handle_edit_menu, (gpointer)6, NULL,
+    { GNOME_APP_UI_TOGGLEITEM, N_("_Mark mode"), NULL, menubar_mark_mode_toggled, NULL, NULL,
       GNOME_APP_PIXMAP_NONE, 0, 'B', GDK_CONTROL_MASK, NULL },
-    { GNOME_APP_UI_ITEM, N_("C_ut"), NULL, menubar_handle_edit_menu, (gpointer)7, NULL,
+    { GNOME_APP_UI_ITEM, N_("C_lear block marks"), NULL, menubar_handle_edit_menu, (gpointer)9, NULL,
+      GNOME_APP_PIXMAP_NONE, 0, 'B', GDK_CONTROL_MASK | GDK_SHIFT_MASK, NULL },
+    { GNOME_APP_UI_ITEM, N_("C_ut"), NULL, menubar_handle_edit_menu, (gpointer)10, NULL,
       GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_CUT, 'X', GDK_CONTROL_MASK, NULL },
-    { GNOME_APP_UI_ITEM, N_("_Copy"), NULL, menubar_handle_edit_menu, (gpointer)8, NULL,
+    { GNOME_APP_UI_ITEM, N_("_Copy"), NULL, menubar_handle_edit_menu, (gpointer)11, NULL,
       GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_COPY, 'C', GDK_CONTROL_MASK, NULL },
-    { GNOME_APP_UI_ITEM, N_("_Paste"), NULL, menubar_handle_edit_menu, (gpointer)9, NULL,
+    { GNOME_APP_UI_ITEM, N_("_Paste"), NULL, menubar_handle_edit_menu, (gpointer)12, NULL,
       GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PASTE, 'V', GDK_CONTROL_MASK, NULL },
+    { GNOME_APP_UI_ITEM, N_("_Interpolate effects"), NULL, menubar_handle_edit_menu, (gpointer)8, NULL,
+      GNOME_APP_PIXMAP_NONE, 0, 'I', GDK_CONTROL_MASK, NULL },
 
     GNOMEUIINFO_END
+
 };
 
 static GnomeUIInfo edit_menu[] = {
@@ -307,6 +377,28 @@ static GnomeUIInfo edit_menu[] = {
     GNOMEUIINFO_END
 };
 
+static GnomeUIInfo track_editor_popup_edit_selection_menu[] = {
+    { GNOME_APP_UI_ITEM, N_("C_lear block marks"), NULL, menubar_handle_edit_menu, (gpointer)9, NULL,
+      GNOME_APP_PIXMAP_NONE, 0, 'B', GDK_CONTROL_MASK | GDK_SHIFT_MASK, NULL },
+    { GNOME_APP_UI_ITEM, N_("C_ut"), NULL, menubar_handle_edit_menu, (gpointer)10, NULL,
+      GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_CUT, 'X', GDK_CONTROL_MASK, NULL },
+    { GNOME_APP_UI_ITEM, N_("_Copy"), NULL, menubar_handle_edit_menu, (gpointer)11, NULL,
+      GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_COPY, 'C', GDK_CONTROL_MASK, NULL },
+    { GNOME_APP_UI_ITEM, N_("_Paste"), NULL, menubar_handle_edit_menu, (gpointer)12, NULL,
+      GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PASTE, 'V', GDK_CONTROL_MASK, NULL },
+    { GNOME_APP_UI_ITEM, N_("_Interpolate effects"), NULL, menubar_handle_edit_menu, (gpointer)8, NULL,
+      GNOME_APP_PIXMAP_NONE, 0, 'I', GDK_CONTROL_MASK, NULL },
+
+    GNOMEUIINFO_END
+};
+
+GnomeUIInfo track_editor_popup_menu[] =	{
+    GNOMEUIINFO_SUBTREE (N_("_Pattern"), edit_pattern_menu),
+    GNOMEUIINFO_SUBTREE (N_("_Track"), edit_track_menu),
+    GNOMEUIINFO_SUBTREE (N_("_Selection"), track_editor_popup_edit_selection_menu),
+    GNOMEUIINFO_END
+};
+
 static GnomeUIInfo pattern_menu[] = {
     { GNOME_APP_UI_ITEM, N_("_Find Unused Pattern"), NULL, modinfo_find_unused_pattern, (gpointer)0, NULL,
       GNOME_APP_PIXMAP_NONE, 0, 'F', GDK_CONTROL_MASK, NULL },
@@ -339,12 +431,24 @@ static GnomeUIInfo instrument_menu[] = {
     GNOMEUIINFO_END
 };
 
+static GnomeUIInfo settings_tracker_menu[] = {
+    { GNOME_APP_UI_TOGGLEITEM, N_("Use _Backing Store"), 0, menubar_backing_store_toggled, 0, 0,
+      GNOME_APP_PIXMAP_NONE, 0, 0, 0, NULL },
+
+    { GNOME_APP_UI_ITEM, N_("_Previous font"), NULL, menubar_settings_tracker_prev_font, (gpointer)0, NULL,
+      GNOME_APP_PIXMAP_NONE, 0, GDK_KP_Subtract, GDK_CONTROL_MASK, NULL },
+    { GNOME_APP_UI_ITEM, N_("_Next font"), NULL, menubar_settings_tracker_next_font, (gpointer)0, NULL,
+      GNOME_APP_PIXMAP_NONE, 0, GDK_KP_Add, GDK_CONTROL_MASK, NULL },
+
+    GNOMEUIINFO_END
+};
+
 // Take care to adjust the indices in menubar_init_prefs() when reordering stuff here!
 static GnomeUIInfo settings_menu[] = {
     { GNOME_APP_UI_TOGGLEITEM, N_("Display _Oscilloscopes"), 0, menubar_scopes_toggled, 0, 0,
       GNOME_APP_PIXMAP_NONE, 0, 0, 0, NULL },
-    { GNOME_APP_UI_TOGGLEITEM, N_("Use _Backing Store"), 0, menubar_backing_store_toggled, 0, 0,
-      GNOME_APP_PIXMAP_NONE, 0, 0, 0, NULL },
+
+    GNOMEUIINFO_SUBTREE (N_("_Tracker"), settings_tracker_menu),
 
     GNOMEUIINFO_SEPARATOR,
 
@@ -352,10 +456,14 @@ static GnomeUIInfo settings_menu[] = {
       GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PREF, 0, 0, NULL },
     { GNOME_APP_UI_ITEM, N_("_Audio Configuration..."), NULL, audioconfig_dialog, NULL, NULL,
       GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PREF, 0, 0, NULL },
-    { GNOME_APP_UI_ITEM, N_("_Playback Effects..."), NULL, effectsconfig_dialog, NULL, NULL,
-      GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PREF, 0, 0, NULL },
     { GNOME_APP_UI_ITEM, N_("_GUI Configuration..."), NULL, gui_settings_dialog, NULL, NULL,
       GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PREF, 0, 0, NULL },
+    /* Define a constant for the index of the MIDI settings item.
+       Will allow to deactivate (grey out) this menu item 
+       if MIDI is not supported. See menubar_create(). */
+#define SETTINGS_MENU_MIDI_INDEX 6
+    { GNOME_APP_UI_ITEM, N_("_MIDI Configuration..."), NULL, midi_settings_dialog, NULL, NULL,
+      GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PREF, 0, 0, NULL },
 
     GNOMEUIINFO_SEPARATOR,
 
@@ -392,21 +500,28 @@ static GnomeUIInfo main_menu[] = {
     GNOMEUIINFO_END
 };
 
+
 void
 menubar_create (GtkWidget *window, GtkWidget *destbox)
 {
-    gnome_app_create_menus(GNOME_APP(window), main_menu);
+    gnome_app_create_menus(GNOME_APP(window), main_menu);	
+
 #ifdef NO_AUDIOFILE
     gtk_widget_set_sensitive(file_menu[SAVE_MOD_AS_WAV_POSITION].widget, FALSE);
 #endif
+
+#if !defined(DRIVER_ALSA_050)
+    gtk_widget_set_sensitive(settings_menu[SETTINGS_MENU_MIDI_INDEX].widget,
+			     FALSE);
+#endif	
 }
 
 void
 menubar_init_prefs ()
 {
-    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(settings_menu[0].widget), gui_settings.gui_display_scopes);
-    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(settings_menu[1].widget), gui_settings.gui_use_backing_store);
-    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(settings_menu[9].widget), gui_settings.save_settings_on_exit);
+  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(settings_menu[0].widget), gui_settings.gui_display_scopes);
+  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(settings_tracker_menu[0].widget), gui_settings.gui_use_backing_store);
+  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(settings_menu[9].widget), gui_settings.save_settings_on_exit);
 }
 
 void
@@ -423,6 +538,7 @@ static GtkItemFactoryEntry menu_items[] = {
     { N_("/File/Save _as..."),                          "<control>A",  fileops_open_dialog, DIALOG_SAVE_MOD },
     { N_("/File/-"),                                    NULL,          NULL, 0, "<Separator>" },
     { N_("/File/Save Module as _WAV..."),               NULL,          fileops_open_dialog, DIALOG_SAVE_MOD_AS_WAV },
+    { N_("/File/Save XM without samples..."),           NULL,          fileops_open_dialog, DIALOG_SAVE_SONG_AS_XM },
     { N_("/File/-"),                                    NULL,          NULL, 0, "<Separator>" },
     { N_("/File/_Quit"),                                "<control>Q",  menubar_quit_requested, 0 },
     { N_("/_Module"),                                   NULL,          NULL, 0, "<Branch>" },
@@ -442,13 +558,18 @@ static GtkItemFactoryEntry menu_items[] = {
     { N_("/Edit/Track/C_ut"),                           "<shift>F3",   menubar_handle_edit_menu, 3 },
     { N_("/Edit/Track/_Copy"),                          "<shift>F4",   menubar_handle_edit_menu, 4 },
     { N_("/Edit/Track/_Paste"),                         "<shift>F5",   menubar_handle_edit_menu, 5 },
-    { N_("/Edit/Track/_Insert"),                        NULL,          menubar_handle_edit_menu, 10 },
-    { N_("/Edit/Track/_Delete"),                        NULL,          menubar_handle_edit_menu, 11 },
+    { N_("/Edit/Track/_Kill notes"),                    "<control>K",  menubar_handle_edit_menu, 13 },
+    { N_("/Edit/Track/_Insert"),                        NULL,          menubar_handle_edit_menu, 6 },
+    { N_("/Edit/Track/_Delete"),                        NULL,          menubar_handle_edit_menu, 7 },
     { N_("/Edit/_Selection"),                           NULL,          NULL, 0, "<Branch>" },
-    { N_("/Edit/Selection/_Mark"),                      "<control>B",  menubar_handle_edit_menu, 6 },
-    { N_("/Edit/Selection/C_ut"),                       "<control>X",  menubar_handle_edit_menu, 7 },
-    { N_("/Edit/Selection/_Copy"),                      "<control>C",  menubar_handle_edit_menu, 8 },
-    { N_("/Edit/Selection/_Paste"),                     "<control>V",  menubar_handle_edit_menu, 9 },
+    { N_("/Edit/Selection/_Mark mode"),                 "<control>B",  menubar_mark_mode_toggled, 0, "<ToggleItem>" },
+    { N_("/Edit/Selection/C_lear block marks"),         "<control><shift>B",  menubar_handle_edit_menu, 9 },
+    { N_("/Edit/Selection/C_ut"),                       "<control>X",  menubar_handle_edit_menu, 10 },
+    { N_("/Edit/Selection/_Copy"),                      "<control>C",  menubar_handle_edit_menu, 11 },
+    { N_("/Edit/Selection/_Paste"),                     "<control>V",  menubar_handle_edit_menu, 12 },
+    { N_("/Edit/Selection/_Interpolate effects"),       "<control>I",  menubar_handle_edit_menu, 8 },
+    { N_("/Edit/Track/Increment cmd value"),            "<control>=",  menubar_handle_edit_menu, 14 },
+    { N_("/Edit/Track/Decrement cmd value"),            "<control>-",  menubar_handle_edit_menu, 15 },
     { N_("/_Pattern"),                                  NULL,          NULL, 0, "<Branch>" },
     { N_("/Pattern/_Find Unused Pattern"),              "<control>F",  modinfo_find_unused_pattern, 0 },
     { N_("/Pattern/_Copy Current to Unused Pattern"),   "<control>G",  modinfo_copy_to_unused_pattern, 0 },
@@ -463,12 +584,15 @@ static GtkItemFactoryEntry menu_items[] = {
     { N_("/Instrument/_Delete Unused Instruments"),     NULL,          modinfo_delete_unused_instruments, 0 },
     { N_("/_Settings"),                                 NULL,          NULL, 0, "<Branch>" },
     { N_("/Settings/Display _Oscilloscopes"),           NULL,          menubar_scopes_toggled, 0, "<ToggleItem>" },
-    { N_("/Settings/Use _Backing Store"),               NULL,          menubar_backing_store_toggled, 0, "<ToggleItem>" },
+    { N_("/Settings/_Tracker"),                         NULL,          NULL, 0, "<Branch>" },
+    { N_("/Settings/Tracker/Use _Backing Store"),       NULL,          menubar_backing_store_toggled, 0, "<ToggleItem>" },
+    { N_("/Settings/Tracker/_Previous font"),           "<control>KP_Subtract", menubar_settings_tracker_prev_font },
+    { N_("/Settings/Tracker/_Next font"),               "<control>KP_Add", menubar_settings_tracker_next_font },
     { N_("/Settings/-"),                                NULL,          NULL, 0, "<Separator>" },
     { N_("/Settings/_Keyboard Configuration..."),       NULL,          keys_dialog, 0 },
     { N_("/Settings/_Audio Configuration..."),          NULL,          audioconfig_dialog, 0 },
-    { N_("/Settings/_Playback Effects..."),             NULL,          effectsconfig_dialog, 0 },
     { N_("/Settings/_GUI Configuration..."),            NULL,          gui_settings_dialog, 0 },
+    { N_("/Settings/_MIDI Configuration..."),           NULL,          midi_settings_dialog, 0 },
     { N_("/Settings/-"),                                NULL,          NULL, 0, "<Separator>" },
     { N_("/Settings/_Save Settings now"),               NULL,          menubar_save_settings_now, 0 },
     { N_("/Settings/Save Settings on _Exit"),           NULL,          menubar_save_settings_on_exit_toggled, 0, "<ToggleItem>" },
@@ -485,6 +609,20 @@ id (const char *p)
     return (char*)p;
 }
 
+static void
+strip_underscores (char *dest, const char *src)
+{
+    char c;
+
+    while((c = *src++)) {
+	if(c != '_') {
+	    *dest++ = c;
+	}
+    }
+
+    *dest = '\0';
+}
+
 void
 menubar_init_prefs ()
 {
@@ -496,16 +634,20 @@ menubar_init_prefs ()
     char * (*tl)(const char *) = id;
 #endif
 
-    prefs_check_prefs_dir();
-    sprintf(buf, "%s/non-gnome-accels", prefs_get_prefsdir());
+    gtk_item_factory_parse_rc(prefs_get_filename("non-gnome-accels"));
 
-    gtk_item_factory_parse_rc(buf);
-
-    w = gtk_item_factory_get_widget(item_factory, tl("/Settings/Display Oscilloscopes"));
+    /* UGH: Oh, oh... gtk+ requires that gtk_item_factory_get_widget()
+       be called without the underscores in the menu string */
+    strip_underscores(buf, tl("/Settings/Display _Oscilloscopes"));
+    w = gtk_item_factory_get_widget(item_factory, buf);
     gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), gui_settings.gui_display_scopes);
-    w = gtk_item_factory_get_widget(item_factory, tl("/Settings/Use Backing Store"));
+
+    strip_underscores(buf, tl("/Settings/Tracker/Use _Backing Store"));
+    w = gtk_item_factory_get_widget(item_factory, buf);
     gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), gui_settings.gui_use_backing_store);
-    w = gtk_item_factory_get_widget(item_factory, tl("/Settings/Save Settings on Exit"));
+
+    strip_underscores(buf, tl("/Settings/Save Settings on _Exit"));
+    w = gtk_item_factory_get_widget(item_factory, buf);
     gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), gui_settings.save_settings_on_exit);
 }
 
@@ -513,7 +655,9 @@ void
 menubar_create (GtkWidget *window, GtkWidget *destbox)
 {
     GtkWidget *thing;
+#ifdef ENABLE_NLS
     int i;
+#endif
 
     accel_group = gtk_accel_group_new ();
 
@@ -521,20 +665,38 @@ menubar_create (GtkWidget *window, GtkWidget *destbox)
 					 "<main>",
 					 accel_group);
 
-    for(i = 0, all_translated = TRUE;
-	all_translated && i < sizeof(menu_items) / sizeof(menu_items[0]); i++) {
-	all_translated &= !strcmp(menu_items[i].path, _(menu_items[i].path));
+    /*
+
+      We need to check whether all menu entries have been
+      translated. If this is not the case, there will be a "File" menu
+      and a menu with a title corresponding to the translation of
+      "File", for example, if only parts of the "File" menu have been
+      translated. Since this looks rather messy, we disable
+      translation of the menu bar completely.
+
+     */
+
+#ifdef ENABLE_NLS
+    for(i = 0, all_translated = TRUE; all_translated && i < sizeof(menu_items) / sizeof(menu_items[0]); i++) {
+        /* as long as all strings have been translated, keep all_translated = TRUE */
+        char *str = strchr(menu_items[i].path + 1, '/');
+        if(str != NULL && strlen(str) > 2) {
+            gboolean string_translated = (0 != strcmp(menu_items[i].path, _(menu_items[i].path)));
+            all_translated &= string_translated;
+            if(!string_translated) {
+                fprintf(stderr, "*** Untranslated menu entry: %s\n", menu_items[i].path);
+            }
+        }
     }
 
     if(all_translated) {
-	for(i = 0; i < sizeof(menu_items) / sizeof(menu_items[0]); i++) {
-	    menu_items[i].path = _(menu_items[i].path);
-	}
+        for(i = 0; i < sizeof(menu_items) / sizeof(menu_items[0]); i++) {
+            menu_items[i].path = _(menu_items[i].path);
+        }
     } else {
-#ifdef ENABLE_NLS
-	fprintf(stderr, "*** SoundTracker Non-GNOME: Disabling menu bar translation because\n*** not all strings have been translated.\n");
-#endif
+        fprintf(stderr, "*** SoundTracker Non-GNOME: Disabling menu bar translation because\n*** not all strings have been translated.\n*** ( --- this is NOT the user's fault! :D --- )\n");
     }
+#endif
 
     gtk_item_factory_create_items (item_factory,
 				   sizeof(menu_items) / sizeof(menu_items[0]),
@@ -550,12 +712,35 @@ menubar_create (GtkWidget *window, GtkWidget *destbox)
 void
 menubar_write_accels ()
 {
-    char buf[256];
-
-    prefs_check_prefs_dir();
-    sprintf(buf, "%s/non-gnome-accels", prefs_get_prefsdir());
-
-    gtk_item_factory_dump_rc(buf, NULL, TRUE);
+    gtk_item_factory_dump_rc(prefs_get_filename("non-gnome-accels"), NULL, TRUE);
 }
 
 #endif
+
+void
+menubar_block_mode_set(gboolean state)
+{
+#ifdef USE_GNOME
+
+    mark_mode_toggle_ignore = TRUE;
+    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(edit_selection_menu[0].widget), state);
+
+#else /* USE_GNOME */	
+    GtkWidget *w;
+    char buf[256];
+#ifdef ENABLE_NLS
+    char * (*tl)(const char *) = all_translated ? gettext : id;
+#else
+    char * (*tl)(const char *) = id;
+#endif
+
+    mark_mode_toggle_ignore = TRUE;
+
+    strip_underscores(buf, tl("/Edit/Selection/_Mark mode"));
+    w = gtk_item_factory_get_widget(item_factory, buf);
+    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), state);
+    
+#endif /* USE_GNOME */
+
+    mark_mode_toggle_ignore = FALSE;
+}
