@@ -2,7 +2,7 @@
 /*
  * The Real SoundTracker - GUI (menu bar) (header)
  *
- * Copyright (C) 1999 Michael Krause
+ * Copyright (C) 1999-2001 Michael Krause
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,6 +22,8 @@
 #ifndef _ST_MENUBAR_H
 #define _ST_MENUBAR_H
 
+#include <config.h>
+
 #include <gtk/gtk.h>
 
 void
@@ -36,4 +38,11 @@ menubar_init_prefs ();
 void
 menubar_write_accels ();
 
+void
+menubar_block_mode_set(gboolean state);
+
+#ifdef USE_GNOME
+extern GnomeUIInfo track_editor_popup_menu[];
+#endif
+
 #endif /* ST_MENUBAR_H */
