@@ -2,7 +2,7 @@
 /*
  * The Real SoundTracker - GTK+ sample display widget
  *
- * Copyright (C) 1998-1999 Michael Krause
+ * Copyright (C) 1998-2001 Michael Krause
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,8 +24,9 @@
 #include "sample-display.h"
 
 #include <gtk/gtksignal.h>
+#include <gtk/gtkmain.h>
 
-#define XPOS_TO_OFFSET(x) (s->win_start + (x) * s->win_length / s->width)
+#define XPOS_TO_OFFSET(x) (s->win_start + ((guint64)(x)) * s->win_length / s->width)
 #define OFFSET_RANGE(l, x) (x < 0 ? 0 : (x >= l ? l - 1 : x))
 
 static const int default_colors[] = {
@@ -56,6 +57,20 @@ enum {
 
 static guint sample_display_signals[LAST_SIGNAL] = { 0 };
 
+static int  sample_display_startoffset_to_xpos (SampleDisplay *s,
+						int offset);
+static gint sample_display_idle_draw_function  (SampleDisplay *s);
+
+static void
+sample_display_idle_draw (SampleDisplay *s)
+{
+    if(!s->idle_handler) {
+	s->idle_handler = gtk_idle_add((GtkFunction)sample_display_idle_draw_function,
+				       (gpointer)s);
+	g_assert(s->idle_handler != 0);
+    }
+}
+
 void
 sample_display_enable_zero_line (SampleDisplay *s,
 				 gboolean enable)
@@ -63,7 +78,6 @@ sample_display_enable_zero_line (SampleDisplay *s,
     s->display_zero_line = enable;
 
     if(s->datalen) {
-	s->complete_redraw = 1;
 	gtk_widget_queue_draw(GTK_WIDGET(s));
     }
 }
@@ -118,9 +132,8 @@ sample_display_set_data (SampleDisplay *s,
     s->selecting = 0;
     
     s->loop_start = -1;
-    
-    s->complete_redraw = 1;
-    gtk_widget_queue_draw(GTK_WIDGET(s));
+
+    gtk_widget_queue_draw(GTK_WIDGET(s));    
 }
 
 void
@@ -159,7 +172,6 @@ sample_display_set_loop (SampleDisplay *s,
     s->loop_start = start;
     s->loop_end = end;
 
-    s->complete_redraw = 1;
     gtk_widget_queue_draw(GTK_WIDGET(s));
     gtk_signal_emit(GTK_OBJECT(s), sample_display_signals[SIG_LOOP_CHANGED], start, end);
 }
@@ -182,7 +194,7 @@ sample_display_set_selection (SampleDisplay *s,
     s->sel_start = start;
     s->sel_end = end;
 
-    gtk_widget_queue_draw(GTK_WIDGET(s));
+    sample_display_idle_draw(s);
     gtk_signal_emit(GTK_OBJECT(s), sample_display_signals[SIG_SELECTION_CHANGED], start, end);
 }
 
@@ -198,11 +210,10 @@ sample_display_set_mixer_position (SampleDisplay *s,
 
     if(offset != s->mixerpos) {
 	s->mixerpos = offset;
-	gtk_widget_queue_draw(GTK_WIDGET(s));
+	sample_display_idle_draw(s);
     }
 }
 
-
 void
 sample_display_set_window (SampleDisplay *s,
 			   int start,
@@ -218,7 +229,6 @@ sample_display_set_window (SampleDisplay *s,
     s->win_length = end - start;
     gtk_signal_emit(GTK_OBJECT(s), sample_display_signals[SIG_WINDOW_CHANGED], start, end);
 
-    s->complete_redraw = 1;
     gtk_widget_queue_draw(GTK_WIDGET(s));
 }
 
@@ -367,7 +377,7 @@ static int
 sample_display_startoffset_to_xpos (SampleDisplay *s,
 				    int offset)
 {
-    int d = offset - s->win_start;
+    gint64 d = offset - s->win_start;
 
     if(d < 0)
 	return 0;
@@ -384,7 +394,7 @@ sample_display_endoffset_to_xpos (SampleDisplay *s,
     if(s->win_length < s->width) {
 	return sample_display_startoffset_to_xpos(s, offset);
     } else {
-	int d = offset - s->win_start;
+	gint64 d = offset - s->win_start;
 	int l = (1 - s->win_length) / s->width;
 
 	/* you get these tests by setting the complete formula below equal to 0 or s->width, respectively,
@@ -478,24 +488,37 @@ sample_display_draw_main (GtkWidget *widget,
 
 	if(s->mixerpos != -1) {
 	    sample_display_do_marker_line(widget->window, s, 0, s->mixerpos, x_min, x_max, s->mixerpos_gc);
+	    s->old_mixerpos = s->mixerpos;
 	}
     }
 }
 
 static void
-sample_display_draw (GtkWidget *widget,
-		     GdkRectangle *area)
+sample_display_draw_update (GtkWidget *widget,
+			    GdkRectangle *area)
 {
     SampleDisplay *s = SAMPLE_DISPLAY(widget);
     GdkRectangle area2 = { 0, 0, 0, s->height };
-    int i, x;
+    int x, i;
     const int x_min = area->x;
     const int x_max = area->x + area->width;
+    gboolean special_draw = FALSE;
 
-    if(s->complete_redraw) {
-	s->complete_redraw = 0;
-	sample_display_draw_main(widget, area);
-	return;
+    if(s->mixerpos != s->old_mixerpos) {
+	/* Redraw area of old position, redraw area of new position. */
+	for(i = 0; i < 2; i++) {
+	    if(s->old_mixerpos >= s->win_start && s->old_mixerpos < s->win_start + s->win_length) {
+		x = sample_display_startoffset_to_xpos(s, s->old_mixerpos);
+		area2.x = MIN(x_max - 1, MAX(x_min, x - 3));
+		area2.width = 7;
+		if(area2.x + area2.width > x_max) {
+		    area2.width = x_max - area2.x;
+		}
+		sample_display_draw_main(widget, &area2);
+	    }
+	    s->old_mixerpos = s->mixerpos;
+	}
+	special_draw = TRUE;
     }
 
     if(s->sel_start != s->old_ss || s->sel_end != s->old_se) {
@@ -535,21 +558,21 @@ sample_display_draw (GtkWidget *widget,
 
 	s->old_ss = s->sel_start;
 	s->old_se = s->sel_end;
+	special_draw = TRUE;
     }
 
-    if(s->mixerpos != s->old_mixerpos) {
-	for(i = 0; i < 2; i++) {
-	    if(s->old_mixerpos >= s->win_start && s->old_mixerpos < s->win_start + s->win_length) {
-		x = sample_display_startoffset_to_xpos(s, s->old_mixerpos);
-		area2.x = MIN(x_max - 1, MAX(x_min, x - 3));
-		area2.width = 7;
-		sample_display_draw_main(widget, &area2);
-	    }
-	    s->old_mixerpos = s->mixerpos;
-	}
+    if(!special_draw) {
+	sample_display_draw_main(widget, area);
     }
 }
 
+static void
+sample_display_draw (GtkWidget *widget,
+		     GdkRectangle *area)
+{
+    sample_display_draw_main(widget, area);
+}
+
 static gint
 sample_display_expose (GtkWidget *widget,
 		       GdkEventExpose *event)
@@ -558,6 +581,20 @@ sample_display_expose (GtkWidget *widget,
     return FALSE;
 }
 
+static gint
+sample_display_idle_draw_function (SampleDisplay *s)
+{
+    GdkRectangle area = { 0, 0, s->width, s->height };
+
+    if(GTK_WIDGET_MAPPED(GTK_WIDGET(s))) {
+	sample_display_draw_update(GTK_WIDGET(s), &area);
+    }
+
+    gtk_idle_remove(s->idle_handler);
+    s->idle_handler = 0;
+    return TRUE;
+}
+
 static void
 sample_display_handle_motion (SampleDisplay *s,
 			      int x,
@@ -644,15 +681,14 @@ sample_display_handle_motion (SampleDisplay *s,
     if(s->sel_start != ss || s->sel_end != se) {
 	s->sel_start = ss;
 	s->sel_end = se;
-	gtk_widget_queue_draw(GTK_WIDGET(s));
+	sample_display_idle_draw(s);
 	gtk_signal_emit(GTK_OBJECT(s), sample_display_signals[SIG_SELECTION_CHANGED], ss, se);
     }
 
     if(s->loop_start != ls || s->loop_end != le) {
 	s->loop_start = ls;
 	s->loop_end = le;
-	s->complete_redraw = 1;
-	gtk_widget_queue_draw(GTK_WIDGET(s));
+	sample_display_idle_draw(s);
 	gtk_signal_emit(GTK_OBJECT(s), sample_display_signals[SIG_LOOP_CHANGED], ls, le);
     }
 }
@@ -784,19 +820,6 @@ sample_display_motion_notify (GtkWidget *widget,
     return TRUE;
 }
 
-static void
-my_2ints_marshal (GtkObject *object,
-		  GtkSignalFunc func,
-		  gpointer func_data,
-		  GtkArg *args)
-{
-    typedef void (*my_2ints_marshal_func)(SampleDisplay *, int, int);
-    my_2ints_marshal_func rfunc = (my_2ints_marshal_func) func;
-    (*rfunc) (SAMPLE_DISPLAY(object),
-	      GTK_VALUE_INT(args[0]),
-	      GTK_VALUE_INT(args[1]));
-}
-  
 static void
 sample_display_class_init (SampleDisplayClass *class)
 {
@@ -822,7 +845,7 @@ sample_display_class_init (SampleDisplayClass *class)
 							     GTK_RUN_FIRST,
 							     object_class->type,
 							     GTK_SIGNAL_OFFSET(SampleDisplayClass, selection_changed),
-							     my_2ints_marshal,
+							     gtk_marshal_NONE__INT_INT,
 							     GTK_TYPE_NONE, 2,
 							     GTK_TYPE_INT,
 							     GTK_TYPE_INT);
@@ -830,7 +853,7 @@ sample_display_class_init (SampleDisplayClass *class)
 							GTK_RUN_FIRST,
 							object_class->type,
 							GTK_SIGNAL_OFFSET(SampleDisplayClass, loop_changed),
-							my_2ints_marshal,
+							gtk_marshal_NONE__INT_INT,
 							GTK_TYPE_NONE, 2,
 							GTK_TYPE_INT,
 							GTK_TYPE_INT);
@@ -838,7 +861,7 @@ sample_display_class_init (SampleDisplayClass *class)
 								 GTK_RUN_FIRST,
 								 object_class->type,
 								 GTK_SIGNAL_OFFSET(SampleDisplayClass, window_changed),
-								 my_2ints_marshal,
+								 gtk_marshal_NONE__INT_INT,
 								 GTK_TYPE_NONE, 2,
 								 GTK_TYPE_INT,
 								 GTK_TYPE_INT);
@@ -861,7 +884,7 @@ sample_display_class_init (SampleDisplayClass *class)
 static void
 sample_display_init (SampleDisplay *s)
 {
-    s->complete_redraw = 0;
+    s->idle_handler = 0;
 }
 
 guint
