@@ -2,7 +2,7 @@
 /*
  * The Real SoundTracker - Transposition dialog (header)
  *
- * Copyright (C) 1999 Michael Krause
+ * Copyright (C) 1999-2001 Michael Krause
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,4 +26,4 @@
 
 void        transposition_dialog                   (void);
 
-#endif /* _ST_GUI_SETTINGS_H */
+#endif /* _ST_TRANSPOSITION_H */
