@@ -2,7 +2,7 @@
 /*
  * The Real SoundTracker - XM support routines
  *
- * Copyright (C) 1998-1999 Michael Krause
+ * Copyright (C) 1998-2001 Michael Krause
  *
  * The XM loader is based on the "maube" 0.10.4 source code, Copyright
  * (C) 1997 by Conrad Parker (not much is left over, though :D)
@@ -22,20 +22,32 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+#include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
+#include <stdlib.h>
+#include <ctype.h>
+#include <dirent.h>
+#include <sys/stat.h>
+#include <math.h>
+#include <sys/wait.h>
 
 #include <glib.h>
 
 #include "i18n.h"
+#include "gui-settings.h"
 #include "xm.h"
+#include "xm-player.h"
 #include "endian-conv.h"
 #include "st-subs.h"
 #include "recode.h"
 #include "errors.h"
 #include "audio.h"
 
+#define LFSTAT_IS_MODULE 1
+
 static guint16 npertab[60]={
     /* -> Tuning 0 */
     1712,1616,1524,1440,1356,1280,1208,1140,1076,1016, 960, 906,
@@ -46,17 +58,33 @@ static guint16 npertab[60]={
 };
 
 void
-byteswap_16_array (gint16 *data,
-		   int count)
+xm_freq_note_to_relnote_finetune (float frequency,
+				  unsigned note,
+				  gint8 *relnote,
+				  gint8 *finetune)
 {
-#ifdef WORDS_BIGENDIAN
-    for(; count; count--, data++) {
-	gint8 *p = (gint16*)data;
-	gint8 a = p[0];
-	p[0] = p[1];
-	p[1] = a;
+    int pitch = - log(frequency / 8363.0) / M_LN2 * (float)PITCH_OCTAVE;
+    int nn = 256 * (49 - note) - pitch;
+    int rn;
+
+    /* nn is now 256*relnote + 2*finetune. Find those two unknowns. */
+    for(rn = 0; nn < -256; rn--) {
+	nn += 256;
     }
-#endif
+    for(; nn > 255; rn++) {
+	nn -= 256;
+    }
+
+    if(rn < -96) {
+	rn = -96;
+    } else if(rn > +96) {
+	rn = +96;
+    }
+
+    g_assert(nn >= -256 && nn <= 255);
+
+    *relnote = rn;
+    *finetune = nn / 2;
 }
 
 static void
@@ -151,19 +179,22 @@ xm_load_xm_pattern (XMPattern *pat,
 		    FILE *f)
 {
     guint8 ph[9];
-    int i, j, len;
+    int i, j;
+    guint16 len;
 
     fread(ph, 1, sizeof(ph), f);
 
     len = get_le_16(ph + 5);
-    if(len > 256 || len < 1) {
+    if(len > 256) {
 	char buf[128];
 	sprintf(buf, _("Pattern length out of range: %d.\n"), len);
 	error_error(buf);
 	return 0;
     }
 
-    if(!st_init_pattern_channels(pat, len, num_channels))
+    if(!st_init_pattern_channels(pat,
+				 len > 0 ? len : 1,
+				 (num_channels + 1) & 0xfe))
 	return 0;
 
     if(get_le_16(ph + 7) == 0)
@@ -193,7 +224,7 @@ xm_load_patterns (XMPattern ptr[],
 	if(i < num_patterns)
 	    e = loadfunc(&ptr[i], num_channels, f);
 	else
-	    e = st_init_pattern_channels(&ptr[i], 64, num_channels);
+	    e = st_init_pattern_channels(&ptr[i], 64, (num_channels + 1) & 0xfe);
 
 	if(!e)
 	    return 0;
@@ -270,31 +301,35 @@ xm_load_xm_samples (STSample samples[],
 	    s->sample.data = NULL;
 	    continue;
 	}
-	s->sample.type = s->sample.looptype & 0x10 ? 16 : 8;
+	s->treat_as_8bit = !(s->sample.looptype & 0x10);
 	s->sample.looptype &= 3;
 
-	if(s->sample.type == 16) {
+	if(!s->treat_as_8bit) {
 	    /* 16 bit sample */
+	    s->treat_as_8bit = FALSE;
 	    s->sample.length >>= 1;
 	    s->sample.loopstart >>= 1;
 	    s->sample.loopend >>= 1;
 
 	    d16 = s->sample.data = malloc(2 * s->sample.length);
 	    fread(d16, 1, 2 * s->sample.length, f);
-	    byteswap_16_array(d16, s->sample.length);
+	    le_16_array_to_host_order(d16, s->sample.length);
 
 	    for(j = s->sample.length, p = 0; j; j--) {
 		p += *d16;
 		*d16++ = p;
 	    }
 	} else {
-	    /* 8 bit sample */
-	    d8 = s->sample.data = malloc(s->sample.length);
+	    s->treat_as_8bit = TRUE;
+
+	    d16 = s->sample.data = malloc(2 * s->sample.length);
+	    d8 = (gint8*)d16;
+	    d8 += s->sample.length;
 	    fread(d8, 1, s->sample.length, f);
 
 	    for(j = s->sample.length, p = 0; j; j--) {
-		p += *d8;
-		*d8++ = p;
+		p += *d8++;
+		*d16++ = p << 8;
 	    }
 	}
 
@@ -324,12 +359,12 @@ xm_save_xm_samples (STSample samples[],
 	/* save sample header */
 	s = &samples[i];
 	memset(sh, 0, sizeof(sh));
-	put_le_32(sh + 0, s->sample.length * (s->sample.type / 8));
-	put_le_32(sh + 4, s->sample.loopstart * (s->sample.type / 8));
-	put_le_32(sh + 8, (s->sample.loopend - s->sample.loopstart) * (s->sample.type / 8));
+	put_le_32(sh + 0, s->sample.length * (s->treat_as_8bit ? 1 : 2));
+	put_le_32(sh + 4, s->sample.loopstart * (s->treat_as_8bit ? 1 : 2));
+	put_le_32(sh + 8, (s->sample.loopend - s->sample.loopstart) * (s->treat_as_8bit ? 1 : 2));
 	sh[12] = s->volume;
 	sh[13] = s->finetune;
-	sh[14] = s->sample.looptype | (s->sample.type == 16 ? 0x10 : 0);
+	sh[14] = s->sample.looptype | (s->treat_as_8bit ? 0x00 : 0x10);
 	sh[15] = s->panning;
 	sh[16] = s->relnote;
 	sh[17] = 0;
@@ -341,7 +376,8 @@ xm_save_xm_samples (STSample samples[],
     for(i = 0; i < num_samples; i++) {
 	s = &samples[i];
 
-	if(s->sample.type == 16) {
+	if(!s->treat_as_8bit) {
+	    // Save as 16 bit sample
 	    gint16 *packbuf, *d16, *ss;
 	    gint16 p, d;
 
@@ -355,20 +391,23 @@ xm_save_xm_samples (STSample samples[],
 		p = *d16++;
 	    }
 
+	    le_16_array_to_host_order(packbuf, s->sample.length);
 	    fwrite(packbuf, 1, s->sample.length * 2, f);
 	    free(packbuf);
 	} else {
-	    gint8 *packbuf, *d8, *ss;
+	    // Save as 8 bit sample
+	    gint16 *d16;
+	    gint8 *packbuf, *ss;
 	    gint8 p, d;
 
 	    packbuf = malloc(s->sample.length);
-	    d8 = s->sample.data;
+	    d16 = s->sample.data;
 	    ss = packbuf;
 
 	    for(k = s->sample.length, p = 0, d = 0; k; k--) {
-		d = *d8 - p;
+		d = (*d16 >> 8) - p;
 		*ss++ = d;
-		p = *d8++;
+		p = (*d16++ >> 8);
 	    }
 
 	    fwrite(packbuf, 1, s->sample.length, f);
@@ -397,27 +436,44 @@ xm_load_xm_instrument (STInstrument *instr,
 		       FILE *f)
 {
     guint8 a[29], b[38];
-    int num_samples;
+    guint16 num_samples;
+    guint32 iheader_size;
 
     st_clean_instrument(instr, NULL);
 
     fread(a, 1, sizeof(a), f);
-
+    iheader_size = get_le_32(a);
     strncpy(instr->name, a + 4, 22);
     recode_ibmpc_to_latin1(instr->name, 22);
+
+    if(iheader_size <= 29) {
+	return 1;
+    }
+
     num_samples = get_le_16(a + 27);
+    if(num_samples > 16) {
+	error_error("XM Load Error: Number of samples in instrument > 16.\n");
+	return 0;
+    }
 
     if(num_samples == 0) {
-	fread(a, 1, 4, f);
+	int to_read = iheader_size - sizeof(a);
+	while(to_read) {
+	    int r = MIN(to_read, sizeof(a));
+	    fread(a, 1, r, f);
+	    to_read -= r;
+	}
     } else {
 	fread(a, 1, 4, f);
 	if(get_le_32(a) != 40) {
-	    error_error("instrument header size != 40\n");
+	    error_error("XM Load Error: Sample header size != 40.\n");
 	    return 0;
 	}
 	fread(instr->samplemap, 1, 96, f);
 	fread(instr->vol_env.points, 1, 48, f);
+	le_16_array_to_host_order((gint16*)instr->vol_env.points, 24);
 	fread(instr->pan_env.points, 1, 48, f);
+	le_16_array_to_host_order((gint16*)instr->pan_env.points, 24);
 
 	fread(b, 1, sizeof(b), f);
 	instr->vol_env.num_points = b[0];
@@ -438,7 +494,7 @@ xm_load_xm_instrument (STInstrument *instr,
 	if(instr->vibtype >= 4) {
 	    char buf[128];
 	    instr->vibtype = 0;
-	    sprintf(buf, _("Invalid vibtype %d, using Sine.\n"), instr->vibtype);
+	    sprintf(buf, "XM Load Warning: Invalid vibtype %d, using Sine.\n", instr->vibtype);
 	    error_warning(buf);
 	}
 	instr->vibrate = b[13];
@@ -486,9 +542,9 @@ xm_load_xi (STInstrument *instr,
 
     fread(instr->samplemap, 1, 96, f);
     fread(instr->vol_env.points, 1, 48, f);
-    byteswap_16_array((gint16*)instr->vol_env.points, 24);
+    le_16_array_to_host_order((gint16*)instr->vol_env.points, 24);
     fread(instr->pan_env.points, 1, 48, f);
-    byteswap_16_array((gint16*)instr->pan_env.points, 24);
+    le_16_array_to_host_order((gint16*)instr->pan_env.points, 24);
 
     fread(b, 1, 16, f);
     instr->vol_env.num_points = b[0];
@@ -547,10 +603,10 @@ xm_save_xi (STInstrument *instr,
 
     fwrite(instr->samplemap, 1, 96, f);
     memcpy(a, instr->vol_env.points, 48);
-    byteswap_16_array((gint16*)a, 24);
+    le_16_array_to_host_order((gint16*)a, 24);
     fwrite(a, 1, 48, f);
     memcpy(a, instr->pan_env.points, 48);
-    byteswap_16_array((gint16*)a, 24);
+    le_16_array_to_host_order((gint16*)a, 24);
     fwrite(a, 1, 48, f);
 
     a[0] = instr->vol_env.num_points;
@@ -581,9 +637,10 @@ xm_save_xi (STInstrument *instr,
 
 static void
 xm_save_xm_instrument (STInstrument *instr,
-		       FILE *f)
+                       FILE *f,
+                       gboolean mode)
 {
-    guint8 h[38];
+    guint8 h[48];
     int num_samples;
 
     num_samples = st_instrument_num_save_samples(instr);
@@ -591,6 +648,10 @@ xm_save_xm_instrument (STInstrument *instr,
     memset(h, 0, sizeof(h));
     strncpy(h + 4, instr->name, 22);
     recode_latin1_to_ibmpc(h + 4, 22);
+
+    if(mode==FALSE)
+	num_samples = 0;
+
     h[27] = num_samples;
 
     if(num_samples == 0) {
@@ -608,8 +669,12 @@ xm_save_xm_instrument (STInstrument *instr,
     fwrite(h, 1, 4, f);
 
     fwrite(instr->samplemap, 1, 96, f);
-    fwrite(instr->vol_env.points, 1, 48, f);
-    fwrite(instr->pan_env.points, 1, 48, f);
+    memcpy(h, instr->vol_env.points, 48);
+    le_16_array_to_host_order((gint16*)h, 24);
+    fwrite(h, 1, 48, f);
+    memcpy(h, instr->pan_env.points, 48);
+    le_16_array_to_host_order((gint16*)h, 24);
+    fwrite(h, 1, 48, f);
 
     memset(&h, 0, 38);
     h[0] = instr->vol_env.num_points;
@@ -629,10 +694,10 @@ xm_save_xm_instrument (STInstrument *instr,
     h[11] = instr->vibsweep;
 	
     put_le_16(h + 14, instr->volfade);
-
+    
     fwrite(&h, 1, 38, f);
 
-    xm_save_xm_samples(instr->samples, f, num_samples);
+    if (mode==TRUE) xm_save_xm_samples(instr->samples, f, num_samples);
 }
 
 static void
@@ -701,7 +766,7 @@ xm_init_locks (XM *xm)
 }
 
 static XM *
-xm_load_mod (FILE *f)
+xm_load_mod (FILE *f,int *status)
 {
     XM *xm;
     guint8 sh[31][8];
@@ -729,7 +794,8 @@ xm_load_mod (FILE *f)
 
     fread(xm->pattern_order_table, 1, 128, f);
     fread(mh, 1, 4, f);
-    
+
+    *status = LFSTAT_IS_MODULE;
     if ((!memcmp("M.K.", mh, 4)) ||
         (!memcmp("M&K!", mh, 4)) ||
         (!memcmp("M!K!", mh, 4)))
@@ -754,7 +820,8 @@ xm_load_mod (FILE *f)
     }
     else
     {
-	error_error(_("No FastTracker XM and no supported MOD format!"));
+        *status &= ~LFSTAT_IS_MODULE;  /* see notes in File_Load() about
+                                          this status*/
 	goto ende;
     }
 
@@ -765,6 +832,8 @@ xm_load_mod (FILE *f)
 
     xm->tempo = 6;
     xm->bpm = 125;
+    player_tempo = xm->tempo;
+    player_bpm = xm->bpm;
     xm->flags = XM_FLAGS_IS_MOD | XM_FLAGS_AMIGA_FREQ;
 
     if(!xm_load_patterns(xm->patterns, n + 1, xm->num_channels, f, xm_load_mod_pattern)) {
@@ -782,7 +851,7 @@ xm_load_mod (FILE *f)
 	    s->volume = sh[i][3];
 	    s->sample.loopstart = get_be_16(sh[i] + 4) << 1;
 	    s->sample.loopend = s->sample.loopstart + (get_be_16(sh[i] + 6) << 1);
-	    s->sample.type = 8;
+	    s->treat_as_8bit = TRUE;
 	    s->panning = 128;
 	    
 	    if(get_be_16(sh[i] + 6) > 1)
@@ -801,12 +870,21 @@ xm_load_mod (FILE *f)
 		s->sample.loopend = 1;
 		s->sample.looptype = 0;
 	    }
-	    
-	    s->sample.data = malloc(s->sample.length);
-	    fread(s->sample.data, 1, s->sample.length, f);
+
+	    s->sample.data = malloc(2 * s->sample.length);
+	    if(!s->sample.data) {
+		goto ende;
+	    }
+	    fread((char*)s->sample.data + s->sample.length, 1, s->sample.length, f);
+	    st_convert_sample((char*)s->sample.data + s->sample.length,
+			      s->sample.data,
+			      8,
+			      16,
+			      s->sample.length);
 	}
     }
 
+    fclose(f);
     return xm;
 
   ende:
@@ -817,16 +895,17 @@ xm_load_mod (FILE *f)
 }
 
 XM *
-XM_Load (const char *filename)
+XM_Load (const char *filename,int *status)
 {
     XM *xm;
     FILE *f;
     guint8 xh[80];
     int i, j, num_patterns, num_instruments;
 
-    f = fopen(filename, "r");
+    *status = 0;
+    f = fopen(filename, "rb");
     if(!f) {
-	error_error(_("Can't open file"));
+        error_error(_("Can't open file"));
 	return NULL;
     }
 
@@ -836,7 +915,7 @@ XM_Load (const char *filename)
        || strncmp(xh + 0, "Extended Module: ", 17) != 0
        || xh[37] != 0x1a) {
 	fseek(f, 0, SEEK_SET);
-	return xm_load_mod(f);
+	return xm_load_mod(f,status);
     }
 
     if(get_le_32(xh + 60) != 276) {
@@ -848,6 +927,9 @@ XM_Load (const char *filename)
 	goto fileerr;
     }
 
+    *status |= LFSTAT_IS_MODULE;  /* see notes in File_Load() about
+                                     this status*/
+
     xm = calloc(1, sizeof(XM));
     if(!xm)
 	goto fileerr;
@@ -857,6 +939,11 @@ XM_Load (const char *filename)
     recode_ibmpc_to_latin1(xm->name, 20);
     xm->song_length = get_le_16(xh + 64);
     xm->restart_position = get_le_16(xh + 66);
+
+    if(xm->restart_position >= xm->song_length) {
+	xm->restart_position = xm->song_length - 1;
+    }
+
     xm->num_channels = get_le_16(xh + 68);
     num_patterns = get_le_16(xh + 70);
     num_instruments = get_le_16(xh + 72);
@@ -865,6 +952,8 @@ XM_Load (const char *filename)
     }
     xm->tempo = get_le_16(xh + 76);
     xm->bpm = get_le_16(xh + 78);
+    player_tempo = xm->tempo;
+    player_bpm = xm->bpm;
     fread(xm->pattern_order_table, 1, 256, f);
 
     if(!xm_load_patterns(xm->patterns, num_patterns, xm->num_channels, f, xm_load_xm_pattern)) {
@@ -891,8 +980,14 @@ XM_Load (const char *filename)
 	    }
 	}
     }
-  weiter:
 
+  weiter:
+    if(xm->num_channels & 1) {
+	/* Yes, mods like these *do* exist. */
+	xm->num_channels++;
+    }
+
+    fclose(f);
     return xm;
     
   ende:
@@ -904,14 +999,15 @@ XM_Load (const char *filename)
 
 int
 XM_Save (XM *xm,
-	 const char *filename)
+	 const char *filename,
+	 gboolean song)
 {
     FILE *f;
     int i;
     guint8 xh[80];
     int num_patterns, num_instruments;
 
-    f = fopen(filename, "w");
+    f = fopen(filename, "wb");
     if(!f)
 	return 0;
 
@@ -941,7 +1037,10 @@ XM_Save (XM *xm,
 	xm_save_xm_pattern(&xm->patterns[i], xm->num_channels, f);
 
     for(i = 0; i < num_instruments; i++)
-	xm_save_xm_instrument(&xm->instruments[i], f);
+        if(song==TRUE)
+            xm_save_xm_instrument(&xm->instruments[i], f, FALSE);
+        else
+            xm_save_xm_instrument(&xm->instruments[i], f, TRUE);
    
     if(ferror(f)) {
 	fclose(f);
@@ -966,6 +1065,8 @@ XM_New ()
     xm->num_channels = 8;
     xm->tempo = 6;
     xm->bpm = 125;
+    player_tempo = xm->tempo;
+    player_bpm = xm->bpm;
     if(!xm_load_patterns(xm->patterns, 0, xm->num_channels, NULL, NULL))
 	goto ende;
 
@@ -986,6 +1087,7 @@ XM_Free (XM *xm)
 
 	for(i = 0; i < sizeof(xm->instruments) / sizeof(xm->instruments[0]); i++) {
 	    STInstrument *ins = &xm->instruments[i];
+	    st_clean_instrument(ins, NULL);
 	    for(j = 0; j < sizeof(ins->samples) / sizeof(ins->samples[0]); j++) {
 		g_mutex_free(ins->samples[j].sample.lock);
 	    }
@@ -994,3 +1096,260 @@ XM_Free (XM *xm)
 	free(xm);
     }
 }
+
+/**************************************************************************
+* Mezzinane compression Handler/Loader
+* coded by Jason Nunn <jsno@downunder.net.au>
+* 7/3/2000
+*
+* Fixed to cope with weird filenames
+* Rob Adamson <r.adamson@hotpop.com>
+* 9-Dec-2000
+*
+* these routines handle modules that are compressed in the known formats-
+* zip, gz, lha, bz2.
+*
+* for simple ones like gz and bz2, it just creates a temp file, and tells
+* XM_Load() to load it before deleting the tmp file.
+*
+* For more complex formats like zip and lha, where they can store multiple
+* files, it's an all-out affair- the compressed file is extracted, and
+* each file is loaded by XM_Load(). The first file it comes across that
+* can be successfully loaded is the one that's accepted as the mod file.
+* After it has loaded a mod, it will remove the directory.
+*
+**************************************************************************/
+char *err_msg = "Bzzzz, error extracting song, aborting operation.";
+
+static char *escape_filename(const char *filename)
+{
+    GString *escaped = g_string_new("");
+    const char *s = filename;
+    char *r;
+
+    g_assert(filename != NULL);
+
+    for (; *s; s ++) {
+	if (strchr("\\\"'*?!# ()&|<>", *s)) {
+	    g_string_append_c(escaped, '\\');
+	}
+	g_string_append_c(escaped, *s);
+    }
+    r = escaped->str;
+    g_string_free(escaped, FALSE);
+
+    return r;
+}
+
+/* Apparently one shouldn't use system() because it doesn't handle
+   the child process getting interrupted. */
+static int ExecuteAndWait(const char *command, const char **argv)
+{
+    int pid, status;
+
+    g_assert(command != NULL);
+
+    pid = fork();
+    switch (pid) {
+    case -1: return -1;
+    case 0: {
+	execv(command, (char **)argv);
+	exit(127);
+    }
+    default:
+	do {
+	    if (waitpid(pid, &status, 0)==-1) {
+		if (errno!=EINTR)
+		    return -1;
+	    } else 
+		return status;
+	} while (1);
+    }
+    return -1;
+}
+
+
+/*
+ * this is for zip style archives, where compression format stores
+ * multiple files
+ */
+static XM *
+File_Extract_Archive (const char *extract_cmd,char *tmp_dir_path,int *status)
+{
+    XM *ret = NULL;
+    register int r;
+    DIR *dp;
+    char str[256];
+    const char *args[4] = {"sh", "-c", NULL, NULL};
+
+    /*
+     * extract archive, and store it in tmp directory
+     */
+    args[2] = extract_cmd;
+    r = ExecuteAndWait ("/bin/sh", args);
+    if ((r == -1) || (r == 127)) {
+        g_snprintf (str, sizeof str, "%s (Err 0)",err_msg);
+        error_error (str);
+        return ret;
+    }
+
+    /*
+     * open directory
+     */
+    dp = opendir (tmp_dir_path);
+    if (dp == NULL) {
+        g_snprintf (str, sizeof str, "%s (Err 1)",err_msg);
+        error_error (str);
+        return ret;
+    }
+
+    /*
+     * read each file name. for each file, try to load it.
+     * the first one that successfully loads- accept it.
+     */
+    for (;;) {
+        struct dirent *ds;
+        struct stat sp;
+
+        ds = readdir (dp);
+        if (ds == NULL)
+            break;
+
+        /*
+         * if not a normal file, then ignore. if get a stat() error,
+         * then don't make a fuss. it's no big deal.
+         */
+        g_snprintf (str, sizeof str, "%s/%s",tmp_dir_path,ds->d_name);
+        if (stat (str,&sp) == 0)
+            if (S_ISREG (sp.st_mode)) {
+                ret = XM_Load (str,status);
+                if (ret != NULL)
+                    break;
+
+            }
+    }
+
+    closedir(dp);
+
+    /*
+     * delete tmp directory
+     */
+    g_snprintf (str, sizeof str, "%s -rf %s",gui_settings.rm_path,tmp_dir_path);
+    r = system (str);
+    if ((r == -1) || (r == 127)) {
+        g_snprintf (str, sizeof str, "%s (Err 2)",err_msg);
+        error_error (str);
+    }
+
+    return ret;
+}
+
+/*
+ * this is for zcat and bunzip2 style archives, where compression format
+ * stores only one file
+ */
+static XM *
+File_Extract_SingleFile (const char *extract_cmd,char *tmp_path,int *status)
+{
+    XM *ret = NULL;
+    register int r;
+    char str[256];
+    const char *args[4] = {"sh", "-c", NULL, NULL};
+
+    /*
+     * extract archive- run cmd
+     */
+    args[2] = extract_cmd;
+    r = ExecuteAndWait ("/bin/sh", args);
+    if ((r == -1) || (r == 127)) {
+        g_snprintf (str, sizeof str, "%s (Err 3)",err_msg);
+        error_error (str);
+        return ret;
+    }
+
+    ret = XM_Load (tmp_path,status);
+    unlink (tmp_path); /*delete tmp file*/
+
+    return ret;
+}
+
+/*
+ * this tests a file extension. if it matches, return 1, if not return 0
+ */
+static int
+f_extension_cmp (const char *filename,char *exten)
+{
+    register int a,b,fs,ks;
+
+    fs = strlen (filename);
+    ks = strlen (exten);
+    if (fs < ks)
+        return 0;
+
+    for (a = fs - ks,b = 0;a < fs;a++) {
+        if (exten[b++] == tolower (filename[a]))
+            continue;
+
+        return 0;
+    }
+
+    return 1;
+}
+
+XM *
+File_Load (const char *filename)
+{
+    char tmp_path[256];
+    char *str = NULL, *filename_esc = NULL;
+    int status = 0;
+    XM *ret = NULL;
+
+    g_assert(filename != NULL);
+
+    filename_esc = escape_filename(filename);
+    tmpnam(tmp_path);
+
+    /* test and load zip files. for unzip version 5.31 */
+    if (f_extension_cmp (filename,".zip")) {
+	str = g_strdup_printf("%s %s -d %s >>/dev/null", gui_settings.unzip_path, filename_esc, tmp_path);
+        ret = File_Extract_Archive (str, tmp_path, &status);
+    }
+
+    /* test and load lha files. for UNIX V1.00   */
+    else if (f_extension_cmp (filename,".lzh") ||
+             f_extension_cmp (filename,".lha")) {
+	str = g_strdup_printf("%s ew=%s %s >>/dev/null", gui_settings.lha_path, tmp_path, filename_esc);
+        ret = File_Extract_Archive (str, tmp_path, &status);
+    }
+
+    /* for zcat 1.2.4 */
+    else if (f_extension_cmp (filename,".gz")) {
+	str = g_strdup_printf("%s %s >> %s", gui_settings.gz_path, filename_esc, tmp_path);
+        ret = File_Extract_SingleFile (str, tmp_path, &status);
+    }
+
+    /* for bunzip2 Version 0.9.0b     */
+    else if (f_extension_cmp (filename,".bz2")) {
+	str = g_strdup_printf("%s -c %s >> %s", gui_settings.bz2_path, filename_esc, tmp_path);
+        ret = File_Extract_SingleFile (str, tmp_path, &status);
+    }
+
+    /* if not compressed, load as normal   */
+    else
+      ret = XM_Load (filename,&status);
+
+    /*
+     * mike: we have this here instead of in xm_load_mod(), because it
+     * can get called multiple times when a compressed file has many
+     * files of unknown types. So, we have to have it here to prevent
+     * lots of dialog boxes popping up with this message, even when a
+     * mod file is eventually found in a compressed archive. -- jsno
+     */
+    if (!(status & LFSTAT_IS_MODULE))
+        error_error (_("Not FastTracker XM and not supported MOD format!"));
+
+    g_free(str);
+    g_free(filename_esc);
+
+    return ret;
+}
